from typing import Dict

from boa3_test.test_drive.model.wallet.account import Account
from boa3_test.test_drive.neoxp.command.neoexpresscommand.oracle.ioraclecommand import IOracleCommand


class OracleEnableCommand(IOracleCommand):
    def __init__(self, account: Account,
                 trace: bool = False,
                 neo_express_data_file: str = None):

        self.trace = trace
        self.input = neo_express_data_file

        command_id = 'enable'
        args = [
            account.get_identifier(),
        ]

        super().__init__(command_id, args)

    def _get_options(self) -> Dict[str, str]:
        options = super()._get_options()

        if self.trace:
            options['--trace'] = ''

        return options



import logging
import os
import threading
from typing import Any, Dict, Optional, Tuple, Union
from unittest import TestCase

from boa3.internal import constants, env
from boa3.internal.analyser.analyser import Analyser
from boa3.internal.compiler.compiler import Compiler
from boa3.internal.exception.CompilerError import CompilerError
from boa3.internal.model.method import Method

_COMPILER_LOCK = threading.RLock()
_LOGGING_LOCK = threading.Lock()

USE_UNIQUE_NAME = False


class BoaTest(TestCase):
    dirname: str = None

    ABORTED_CONTRACT_MSG = 'ABORT is executed'
    ARGUMENT_OUT_OF_RANGE_MSG_PREFIX = 'Specified argument was out of the range of valid values.'
    ASSERT_RESULTED_FALSE_MSG = 'ASSERT is executed with false result.'
    BAD_SCRIPT_EXCEPTION_MSG = "Neo.VM.BadScriptException: Exception of type 'Neo.VM.BadScriptException' was thrown."
    CANT_CALL_SYSCALL_WITH_FLAG_MSG_PREFIX = "Cannot call this SYSCALL with the flag"
    CANT_FIND_METHOD_MSG_PREFIX = "Can't find method"
    CANT_PARSE_VALUE_MSG = "The value could not be parsed."
    CANT_CALL_METHOD_PREFIX = "Cannot Call Method"
    CALLED_CONTRACT_DOES_NOT_EXIST_MSG = 'Called Contract Does Not Exist'
    CONTRACT_NOT_FOUND_MSG_REGEX = 'contract "(.*?)" not found'
    GAS_MUST_BE_POSITIVE_MSG = 'GAS must be positive.'
    GIVEN_KEY_NOT_PRESENT_IN_DICT_MSG_REGEX = "The given key '\\S+' was not present in the dictionary."
    INSUFFICIENT_GAS = 'Insufficient GAS.'
    MAP_KEY_NOT_FOUND_ERROR_MSG = 'Key not found in Map'
    MAX_ITEM_SIZE_EXCEED_MSG_PREFIX = 'MaxItemSize exceed'
    FORMAT_METHOD_DOESNT_EXIST_IN_CONTRACT_MSG_REGEX_PREFIX = r'^Method "{0}" with \d+ parameter\(s\) ' \
                                                              r"doesn't exist in the contract"
    METHOD_DOESNT_EXIST_IN_CONTRACT_MSG_REGEX_PREFIX = FORMAT_METHOD_DOESNT_EXIST_IN_CONTRACT_MSG_REGEX_PREFIX.format(r'\S+')
    NULL_POINTER_MSG = 'Object reference not set to an instance of an object.'
    UNHANDLED_EXCEPTION_MSG_PREFIX = "An unhandled exception was thrown."
    VALUE_CANNOT_BE_NEGATIVE_MSG = 'value can not be negative'
    VALUE_DOES_NOT_FALL_WITHIN_EXPECTED_RANGE_MSG = 'Value does not fall within the expected range.'
    VALUE_IS_OUT_OF_RANGE_MSG_REGEX_SUFFIX = r'The value( \S*)? is out of range.$'

    default_folder: str = ''

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._use_custom_name = USE_UNIQUE_NAME

    @classmethod
    def setUpClass(cls):
        folders = os.path.abspath(__file__).split(os.sep)
        cls.dirname = '/'.join(folders[:-3])
        cls.test_root_dir = '/'.join(folders[-3:-2])
        cls.default_test_folder = ('{0}/{1}'.format(cls.test_root_dir, cls.default_folder)
                                   if len(cls.default_folder) else cls.test_root_dir)

        super(BoaTest, cls).setUpClass()
        constants.COMPILER_VERSION = '_unit_tests_'  # to not change test contract script hashes in different versions

    def method_name(self) -> str:
        return self._testMethodName if hasattr(self, '_testMethodName') else self.id()

    def get_compiler_analyser(self, compiler: Compiler) -> Analyser:
        return compiler._analyser

    def get_all_imported_methods(self, compiler: Compiler) -> Dict[str, Method]:
        from boa3.internal.compiler.filegenerator.filegenerator import FileGenerator
        generator = FileGenerator(compiler.result, compiler._analyser, compiler._entry_smart_contract)
        return {constants.VARIABLE_NAME_SEPARATOR.join(name): value for name, value in generator._methods_with_imports.items()}

    def assertCompilerLogs(self, expected_logged_exception, path) -> Union[bytes, str]:
        output, error_msg = self._assert_compiler_logs_error(expected_logged_exception, path)
        if not issubclass(expected_logged_exception, CompilerError):
            return output
        else:
            # filter to get only the error message, without location information
            import re

            result = re.search('^\\d+:\\d+ - (?P<msg>.*?)\t\\W+\\<.*?\\>', error_msg)
            try:
                return result.group('msg')
            except BaseException:
                return output

    def assertCompilerNotLogs(self, expected_logged_exception, path):
        output, expected_logged = self._get_compiler_log_data(expected_logged_exception, path)
        if len(expected_logged) > 0:
            raise AssertionError(f'{expected_logged_exception.__name__} was logged: "{expected_logged[0].message}"')
        return output

    def _assert_compiler_logs_error(self, expected_logged_exception, path):
        output, expected_logged = self._get_compiler_log_data(expected_logged_exception, path)
        if len(expected_logged) < 1:
            raise AssertionError('{0} not logged'.format(expected_logged_exception.__name__))
        return output, expected_logged[0].message

    def _get_compiler_log_data(self, expected_logged_exception, path, fail_fast=False):
        output = None

        with _LOGGING_LOCK:
            with self.assertLogs() as log:
                from boa3.internal.exception.NotLoadedException import NotLoadedException
                try:
                    output = self.compile(path, fail_fast=fail_fast)
                except NotLoadedException:
                    # when an compiler error is logged this exception is raised.
                    pass

            expected_logged = [exception for exception in log.records
                               if isinstance(exception.msg, expected_logged_exception)]
        return output, expected_logged

    def get_all_compile_log_data(self, path: str, *,
                                 get_errors: bool = True,
                                 get_warnings: bool = False,
                                 fail_fast: bool = False) -> Tuple[list, list]:
        from boa3.internal.exception.CompilerWarning import CompilerWarning

        instance_logs = []
        if get_errors:
            instance_logs.append(CompilerError)
        if get_warnings:
            instance_logs.append(CompilerWarning)

        errors = []
        warnings = []
        _, expected_logged = self._get_compiler_log_data(*instance_logs, path, fail_fast=fail_fast)

        if not get_errors:
            if not get_warnings:
                return errors, warnings
            warnings = [log.msg for log in expected_logged]
        else:
            if get_warnings:
                for log in expected_logged:
                    if isinstance(log.msg, CompilerError):
                        errors.append(log.msg)
                    elif isinstance(log.msg, CompilerWarning):
                        warnings.append(log.msg)
            else:
                errors = [log.msg for log in expected_logged]

        return errors, warnings

    def assertStartsWith(self, first: Any, second: Any):
        if not (hasattr(first, 'startswith') and first.startswith(second)):
            self.fail(f'{first} != {second}')

    def get_dir_path(self, *args: str) -> str:
        type_error_message = 'get_contract_path() takes {0} positional argument but {1} were given'
        num_args = len(args)
        if num_args == 0:
            raise TypeError(type_error_message.format(2, num_args + 1))
        if num_args > 2:
            raise TypeError(type_error_message.format(3, num_args + 1))

        values = [None, env.PROJECT_ROOT_DIRECTORY]
        for index, value in enumerate(reversed(args)):
            values[index] = value

        dir_folder, root_path = values
        return '{0}/{1}'.format(root_path, dir_folder)

    def get_contract_path(self, *args: str) -> str:
        """
        Usages:
            get_contract_path(contract_name)
            get_contract_path(dir_folder, contract_name)
            get_contract_path(root_path, dir_folder, contract_name)
        """
        type_error_message = 'get_contract_path() takes {0} positional argument but {1} were given'
        num_args = len(args)
        if num_args == 0:
            raise TypeError(type_error_message.format(2, num_args + 1))
        if num_args > 3:
            raise TypeError(type_error_message.format(4, num_args + 1))
        if os.path.isabs(args[0]):
            return args[0]

        values = [None, self.default_test_folder, env.PROJECT_ROOT_DIRECTORY]
        for index, value in enumerate(reversed(args)):
            values[index] = value

        contract_name, dir_folder, root_path = values

        from os import path
        if not path.exists(dir_folder) and root_path is env.PROJECT_ROOT_DIRECTORY:
            path_folder = '{0}/{1}'.format(root_path, dir_folder)
            if not path.exists(path_folder) and not dir_folder.startswith(self.test_root_dir):
                path_folder = '{0}/{1}/{2}'.format(root_path, self.test_root_dir, dir_folder)
                if not path.exists(path_folder):
                    path_folder = '{0}/{1}/{2}'.format(root_path, self.default_test_folder, dir_folder)

            dir_folder = path_folder
        else:
            if path.exists(dir_folder):
                dir_folder = os.path.abspath(dir_folder)
            else:
                dir_folder = '{0}/{1}'.format(root_path, dir_folder)

        if not contract_name.endswith('.py'):
            contract_name = contract_name + '.py'

        path = '{0}/{1}'.format(dir_folder, contract_name)
        if not os.path.isfile(path):
            raise FileNotFoundError(path)
        else:
            path = os.path.abspath(path).replace(os.path.sep, constants.PATH_SEPARATOR)
        return path

    def get_deploy_file_paths_without_compiling(self, contract_path: str, output_name: str = None) -> Tuple[str, str]:
        if isinstance(output_name, str):
            output_path, output_file = os.path.split(output_name)
            if len(output_path) == 0:
                file_path = os.path.dirname(contract_path)
                file_name, _ = os.path.splitext(os.path.basename(output_name))
                file_path_without_ext = constants.PATH_SEPARATOR.join((file_path, file_name))
            else:
                file_path_without_ext, _ = os.path.splitext(output_name)
        else:
            file_path_without_ext, _ = os.path.splitext(contract_path)

        if USE_UNIQUE_NAME:
            from boa3_test.test_drive import utils
            file_path_without_ext = utils.create_custom_id(file_path_without_ext, use_time=False)

        return f'{file_path_without_ext}.nef', f'{file_path_without_ext}.manifest.json'

    def get_deploy_file_paths(self, *args: str, output_name: str = None,
                              compile_if_found: bool = False, change_manifest_name: bool = False) -> Tuple[str, str]:
        contract_path = self.get_contract_path(*args)
        if isinstance(contract_path, str):
            nef_path, manifest_path = self.get_deploy_file_paths_without_compiling(contract_path, output_name)
            if contract_path.endswith('.py'):
                with _COMPILER_LOCK:
                    if compile_if_found or not (os.path.isfile(nef_path) and os.path.isfile(manifest_path)):
                        # both .nef and .manifest.json are required to execute the smart contract
                        self.compile_and_save(contract_path, output_name=nef_path, log=True,
                                              change_manifest_name=change_manifest_name,
                                              use_unique_name=False  # already using unique name
                                              )

            return nef_path, manifest_path

        return contract_path, contract_path

    def compile(self, path: str, root_folder: str = None, fail_fast: bool = False) -> bytes:
        from boa3.boa3 import Boa3

        with _COMPILER_LOCK:
            result = Boa3.compile(path, root_folder=root_folder, fail_fast=fail_fast,
                                  log_level=logging.getLevelName(logging.INFO)
                                  )

        return result

    def compile_and_save(self, path: str, root_folder: str = None, debug: bool = False, log: bool = True,
                         output_name: str = None, env: str = None, **kwargs) -> Tuple[bytes, Dict[str, Any]]:

        if output_name is not None:
            output_dir, manifest_name = os.path.split(output_name)  # get name
            if len(output_dir) == 0:
                output_dir, _ = os.path.split(path)
                output_name = f'{output_dir}/{manifest_name}'
            manifest_name, _ = os.path.splitext(manifest_name)  # remove extension
        else:
            manifest_name = None

        use_unique_name = kwargs['use_unique_name'] if 'use_unique_name' in kwargs else True
        if not isinstance(output_name, str) or not output_name.endswith('.nef'):
            nef_output, _ = self.get_deploy_file_paths_without_compiling(path)
        elif use_unique_name and USE_UNIQUE_NAME:
            nef_output, _ = self.get_deploy_file_paths_without_compiling(output_name)
        else:
            nef_output = output_name

        if nef_output.endswith('.py'):
            nef_output = nef_output.replace('.py', '.nef')
        manifest_output = nef_output.replace('.nef', '.manifest.json')

        from boa3.boa3 import Boa3
        from boa3.internal.neo.contracts.neffile import NefFile
        with _COMPILER_LOCK:
            Boa3.compile_and_save(path, output_path=nef_output, root_folder=root_folder,
                                  env=env, debug=debug,
                                  show_errors=log, log_level=logging.getLevelName(logging.INFO)
                                  )

        get_raw_nef = kwargs['get_raw_nef'] if 'get_raw_nef' in kwargs else False
        change_manifest_name = kwargs['change_manifest_name'] if 'change_manifest_name' in kwargs else False

        with open(nef_output, mode='rb') as nef:
            file = nef.read()
            if get_raw_nef:
                output = file
            else:
                output = NefFile.deserialize(file).script

        with open(manifest_output) as manifest_file:
            import json
            manifest = json.loads(manifest_file.read())

        if change_manifest_name and manifest_name is not None:
            manifest['name'] = manifest_name
            with _COMPILER_LOCK:
                with open(manifest_output, mode='w') as manifest_file:
                    manifest_file.write(json.dumps(manifest, indent=4))

        return output, manifest

    def get_debug_info(self, path: str) -> Optional[Dict[str, Any]]:
        if path.endswith('.nef'):
            nef_output = path
        else:
            nef_output, _ = self.get_deploy_file_paths_without_compiling(path)
        debug_info_output = nef_output.replace('.nef', '.nefdbgnfo')

        if not os.path.isfile(debug_info_output):
            return None

        from zipfile import ZipFile
        with ZipFile(debug_info_output, 'r') as dbgnfo:
            import json
            debug_info = json.loads(dbgnfo.read(os.path.basename(nef_output.replace('.nef', '.debug.json'))))
        return debug_info

    def get_output(self, path: str, root_folder: str = None) -> Tuple[bytes, Dict[str, Any]]:
        if path.endswith('.nef'):
            nef_output = path
            manifest_output = path.replace('.nef', '.manifest.json')
        else:
            nef_output, manifest_output = self.get_deploy_file_paths_without_compiling(path)

        with _COMPILER_LOCK:
            if not os.path.isfile(nef_output):
                return self.compile_and_save(path, root_folder=root_folder)

        from boa3.internal.neo.contracts.neffile import NefFile

        if not os.path.isfile(nef_output):
            output = bytes()
        else:
            with open(nef_output, mode='rb') as nef:
                file = nef.read()
                output = NefFile.deserialize(file).script

        if not os.path.isfile(manifest_output):
            manifest = {}
        else:
            with open(manifest_output) as manifest_output:
                import json
                manifest = json.loads(manifest_output.read())

        return output, manifest

    def get_bytes_output(self, path: str) -> Tuple[bytes, Dict[str, Any]]:
        nef_output, manifest_output = self.get_deploy_file_paths_without_compiling(path)
        with _COMPILER_LOCK:
            if not os.path.isfile(nef_output):
                return self.compile_and_save(path, get_raw_nef=True)

        if not os.path.isfile(nef_output):
            output = bytes()
        else:
            with open(nef_output, mode='rb') as nef:
                output = nef.read()

        if not os.path.isfile(manifest_output):
            manifest = {}
        else:
            with open(manifest_output) as manifest_output:
                import json
                manifest = json.loads(manifest_output.read())

        return output, manifest



from __future__ import annotations

from typing import List, Dict, Any

from boa3.internal.neo3.core.types import UInt256
from boa3_test.test_drive.model.network.payloads.testblock import TestBlock
from boa3_test.test_drive.model.network.payloads.witness import Witness
from boa3_test.test_drive.model.wallet.account import Account
from boa3_test.test_drive.neoxp import utils
from boa3_test.test_drive.testrunner.blockchain.transaction import TestRunnerTransaction


class TestRunnerBlock(TestBlock):
    def __init__(self):
        self._transactions: List[TestRunnerTransaction] = []
        super().__init__()
        self._hash: UInt256 = UInt256.zero()
        self._size = 0
        self._version = 0
        self._previous_block_hash = UInt256.zero()
        self._merkle_root = UInt256.zero()
        self._nonce = 0
        self._primary_index = 0
        self._next_consensus: Account = None
        self._witnesses: List[Witness] = []

    @property
    def hash(self) -> UInt256:
        return self._hash

    @property
    def size(self) -> int:
        return self._size

    @property
    def version(self) -> int:
        return self._version

    @property
    def previous_block_hash(self) -> UInt256:
        return self._previous_block_hash

    @property
    def merkle_root(self) -> UInt256:
        return self._merkle_root

    @property
    def nonce(self) -> int:
        return self._nonce

    @property
    def primary_index(self) -> int:
        return self._primary_index

    @property
    def next_consensus(self) -> Account:
        return self._next_consensus

    @property
    def witnesses(self) -> List[Witness]:
        return self._witnesses.copy()

    @property
    def transactions(self) -> List[TestRunnerTransaction]:
        return self._transactions.copy()

    def to_json(self) -> Dict[str, Any]:
        json_block = super().to_json()

        # timestamp in this implementation is exported as time
        time = json_block.pop('timestamp')
        json_block['time'] = time

        return json_block

    @classmethod
    def from_json(cls, json: Dict[str, Any], *args, **kwargs) -> TestRunnerBlock:
        from boa3.internal.neo import from_hex_str

        block: TestRunnerBlock = super().from_json(json)

        if 'neoxp_config' in kwargs:
            neoxp_config = kwargs['neoxp_config']
        elif len(args) > 0:
            neoxp_config = args[0]
        else:
            neoxp_config = None

        block._size = int(json['size'])
        block._version = int(json['version'])
        block._previous_block_hash = UInt256(from_hex_str(json['previousblockhash']))
        block._merkle_root = UInt256(from_hex_str(json['merkleroot']))
        block._timestamp = int(json['time'])
        block._nonce = int(json['nonce'], base=16)
        block._primary_index = int(json['primary'])
        next_consensus = json['nextconsensus']
        block._next_consensus = (utils.get_account_from_script_hash_or_id(neoxp_config, next_consensus)
                                 if neoxp_config is not None
                                 else next_consensus
                                 )
        block._witnesses = [Witness.from_json(js) for js in json['witnesses']]
        block._transactions = [TestRunnerTransaction.from_json(js, *args, **kwargs) for js in json['tx']]

        return block



import ast
from typing import Any, Dict, Optional

from boa3.internal import constants
from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class UInt160Method(IBuiltinMethod):

    def __init__(self, return_type: UInt160Type, argument_type: IType = None):
        from boa3.internal.model.type.type import Type
        if argument_type is None or not self.validate_parameters(argument_type):
            argument_type = Type.none

        identifier = 'UInt160'
        args: Dict[str, Variable] = {'object': Variable(argument_type)}

        args_default = ast.parse("{0}".format(Type.int.default_value)
                                 ).body[0].value

        super().__init__(identifier, args, [args_default], return_type=return_type)

    @property
    def _arg_object(self) -> Variable:
        return self.args['object']

    @property
    def identifier(self) -> str:
        from boa3.internal.model.type.type import Type
        if self._arg_object.type is Type.none:
            return self._identifier
        return '-{0}_from_{1}'.format(self._identifier, self._arg_object.type._identifier)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) > 1:
            return False
        if len(params) == 0:
            return True

        from boa3.internal.model.type.itype import IType
        if not isinstance(params[0], (IExpression, IType)):
            return False

        param_type: IType = params[0].type if isinstance(params[0], IExpression) else params[0]
        from boa3.internal.model.type.type import Type

        return (Type.bytes.is_type_of(param_type)
                or Type.int.is_type_of(param_type))

    def evaluate_literal(self, *args: Any) -> Any:
        from boa3.internal.neo3.core.types import UInt160

        if len(args) == 0:
            return UInt160.zero().to_array()

        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, int):
                from boa3.internal.neo.vm.type.Integer import Integer
                arg = Integer(arg).to_byte_array(min_length=UInt160._BYTE_LEN)
            if isinstance(arg, bytes):
                value = UInt160(arg).to_array()
                return value

        return super().evaluate_literal(*args)

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.model.type.type import Type

        code_generator.duplicate_stack_top_item()
        code_generator.insert_type_check(Type.int.stack_item)
        # if isinstance(arg, int):
        is_int = code_generator.convert_begin_if()

        #   assert num >= 0
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(0)
        code_generator.convert_operation(BinaryOp.GtE, is_internal=True)
        code_generator.convert_assert()

        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_literal(constants.SIZE_OF_INT160)
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(2)

        #   if len(num) < 20
        need_to_adjust_size = code_generator.convert_begin_if()
        code_generator.change_jump(need_to_adjust_size, Opcode.JMPGE)

        #       # increase length of number to 20
        code_generator.convert_literal(bytes(constants.SIZE_OF_INT160))
        code_generator.swap_reverse_stack_items(3)
        code_generator.convert_operation(BinaryOp.Sub, is_internal=True)
        code_generator.insert_opcode(Opcode.LEFT, pop_from_stack=True, add_to_stack=[Type.bytes])
        code_generator.convert_operation(BinaryOp.Concat.build(Type.bytes, Type.bytes), is_internal=True)

        #   else:
        need_to_adjust_size = code_generator.convert_begin_else(need_to_adjust_size, is_internal=True)
        #       # clean stack
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()

        code_generator.convert_end_if(need_to_adjust_size, is_internal=True)
        code_generator.convert_end_if(is_int, is_internal=True)

        #   # convert to uint160
        code_generator.convert_cast(self.return_type, is_internal=True)

        # assert len(value) == 20
        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_literal(constants.SIZE_OF_INT160)
        code_generator.convert_operation(BinaryOp.NumEq, is_internal=True)
        code_generator.convert_assert()

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self._arg_object.type):
            return self
        if isinstance(value, list):
            value = value[0] if len(value) > 0 else None
        if self.validate_parameters(value):
            return UInt160Method(self.return_type, value)
        return super().build(value)



from typing import Dict

from boa3_test.test_drive.neoxp.command.neoexpresscommand.neoexpresscommand import NeoExpressCommand

__all__ = ['FastForwardCommand']


class FastForwardCommand(NeoExpressCommand):
    def __init__(self, block_count: int = 0,
                 time_interval_in_secs: int = 0,
                 neo_express_data_file: str = None):

        self.block_count = block_count if block_count > 0 else 0
        self.timestamp_delta = time_interval_in_secs if time_interval_in_secs > 0 else 0
        self.input = neo_express_data_file

        command_id = 'fastfwd'
        args = []
        if isinstance(block_count, int):
            args.append(str(block_count))

        super().__init__(command_id, args)

    def _get_options(self) -> Dict[str, str]:
        options = super()._get_options()

        if self.timestamp_delta > 0:
            options['--timestamp-delta'] = str(self.timestamp_delta)

        return options



__all__ = [
    'Ledger',
]


from typing import List, Optional, Union

from boa3.builtin.interop.blockchain import Block, Signer, Transaction, VMState
from boa3.builtin.type import UInt256, UInt160


class Ledger:
    """
    A class used to represent the Ledger native contract.

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/reference/scapi/framework/native/Ledger>`__
    to learn more about the Ledger class.
    """

    hash: UInt160

    @classmethod
    def get_block(cls, index_or_hash: Union[int, UInt256]) -> Optional[Block]:
        """
        Gets the block with the given index or hash.

        >>> Ledger.get_block(0)        # first block
        {
            'hash': b"S{\\xed'\\x85&\\xf5\\x93U=\\xc1\\xbf'\\x95\\xc4/\\x80X\\xdb\\xd5\\xa1-\\x97q\\x85\\xe3I\\xe5\\x99cd\\x04",
            'version': 0,
            'previous_hash': '\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00',
            'merkle_root': '\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00',
            'timestamp': 1468595301000,
            'nonce': 2083236893,
            'index': 0,
            'primary_index': 0,
            'next_consensus': b'\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04',
            'transaction_count': 0,
        }

        >>> Ledger.get_block(UInt256(b"S{\\xed'\\x85&\\xf5\\x93U=\\xc1\\xbf'\\x95\\xc4/\\x80X\\xdb\\xd5\\xa1-\\x97q\\x85\\xe3I\\xe5\\x99cd\\x04"))        # first block
        {
            'hash': b"S{\\xed'\\x85&\\xf5\\x93U=\\xc1\\xbf'\\x95\\xc4/\\x80X\\xdb\\xd5\\xa1-\\x97q\\x85\\xe3I\\xe5\\x99cd\\x04",
            'version': 0,
            'previous_hash': '\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00',
            'merkle_root': '\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00',
            'timestamp': 1468595301000,
            'nonce': 2083236893,
            'index': 0,
            'primary_index': 0,
            'next_consensus': b'\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04',
            'transaction_count': 0,
        }

        >>> Ledger.get_block(9999999)      # block doesn't exist
        None

        >>> Ledger.get_block(UInt256(bytes(32)))   # block doesn't exist
        None

        :param index_or_hash: index or hash identifier of the block
        :type index_or_hash: int or UInt256
        :return: the desired block, if exists. None otherwise
        :rtype: Block or None
        """
        pass

    @classmethod
    def get_current_index(cls) -> int:
        """
        Gets the index of the current block.

        >>> Ledger.get_current_index()
        10908937

        >>> Ledger.get_current_index()
        2108690

        >>> Ledger.get_current_index()
        3529755

        :return: the index of the current block
        :rtype: int
        """
        pass

    @classmethod
    def get_transaction(cls, hash_: UInt256) -> Optional[Transaction]:
        """
        Gets a transaction with the given hash.

        >>> Ledger.get_transaction(UInt256(b'\\xff\\x7f\\x18\\x99\\x8c\\x1d\\x10X{bA\\xc2\\xe3\\xdf\\xc8\\xb0\\x9f>\\xd0\\xd2G\\xe3\\xba\\xd8\\x96\\xb9\\x0e\\xc1iS\\xcdr'))
        {
            'hash': b'\\xff\\x7f\\x18\\x99\\x8c\\x1d\\x10X{bA\\xc2\\xe3\\xdf\\xc8\\xb0\\x9f>\\xd0\\xd2G\\xe3\\xba\\xd8\\x96\\xb9\\x0e\\xc1iS\\xcdr',
            'version': 0,
            'nonce': 2025056010,
            'sender': b'\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04',
            'system_fee': 2028330,
            'network_fee': 1206580,
            'valid_until_block': 5761,
            'script': b'\\x0c\\x14\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04\\x11\\xc0\\x1f\\x0c\\tbalanceOf\\x0c\\x14\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2Ab}[R',
        }

        >>> Ledger.get_transaction(UInt256(bytes(32)))   # transaction doesn't exist
        None

        :param hash_: hash identifier of the transaction
        :type hash_: UInt256
        :return: the Transaction, if exists. None otherwise
        """
        pass

    @classmethod
    def get_transaction_from_block(cls, block_hash_or_height: Union[UInt256, int], tx_index: int) -> Optional[Transaction]:
        """
        Gets a transaction from a block.

        >>> Ledger.get_transaction_from_block(1, 0)
        {
            'hash': b'\\xff\\x7f\\x18\\x99\\x8c\\x1d\\x10X{bA\\xc2\\xe3\\xdf\\xc8\\xb0\\x9f>\\xd0\\xd2G\\xe3\\xba\\xd8\\x96\\xb9\\x0e\\xc1iS\\xcdr',
            'version': 0,
            'nonce': 2025056010,
            'sender': b'\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04',
            'system_fee': 2028330,
            'network_fee': 1206580,
            'valid_until_block': 5761,
            'script': b'\\x0c\\x14\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04\\x11\\xc0\\x1f\\x0c\\tbalanceOf\\x0c\\x14\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2Ab}[R',
        }

        >>> Ledger.get_transaction_from_block(UInt256(b'\\x21|\\xc2~U\\t\\x89^\\x0c\\xc0\\xd29wl\\x0b\\xad d\\xe1\\xf5U\\xd7\\xf5B\\xa5/\\x8b\\x8f\\x8b\\x22\\x24\\x80'), 0)
        {
            'hash': b'\\xff\\x7f\\x18\\x99\\x8c\\x1d\\x10X{bA\\xc2\\xe3\\xdf\\xc8\\xb0\\x9f>\\xd0\\xd2G\\xe3\\xba\\xd8\\x96\\xb9\\x0e\\xc1iS\\xcdr',
            'version': 0,
            'nonce': 2025056010,
            'sender': b'\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04',
            'system_fee': 2028330,
            'network_fee': 1206580,
            'valid_until_block': 5761,
            'script': b'\\x0c\\x14\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04\\x11\\xc0\\x1f\\x0c\\tbalanceOf\\x0c\\x14\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2Ab}[R',
        }

        >>> Ledger.get_transaction_from_block(123456789, 0)     # height does not exist yet
        None

        >>> Ledger.get_transaction_from_block(UInt256(bytes(32)), 0)     # block hash does not exist
        None

        :param block_hash_or_height: a block identifier
        :type block_hash_or_height: UInt256 or int
        :param tx_index: the transaction identifier in the block
        :type tx_index: int
        :return: the Transaction, if exists. None otherwise
        """
        pass

    @classmethod
    def get_transaction_height(cls, hash_: UInt256) -> int:
        """
        Gets the height of a transaction.

        >>> Ledger.get_transaction_height(UInt256(b'\\x28\\x89\\x4f\\xb6\\x10\\x62\\x9d\\xea\\x4c\\xcd\\x00\\x2e\\x9e\\x11\\xa6\\xd0\\x3d\\x28\\x90\\xc0\\xe5\\xd4\\xfc\\x8f\\xc6\\x4f\\xcc\\x32\\x53\\xb5\\x48\\x01'))
        2108703

        >>> Ledger.get_transaction_height(UInt256(b'\\x29\\x41\\x06\\xdb\\x4c\\xf3\\x84\\xa7\\x20\\x4d\\xba\\x0a\\x04\\x03\\x72\\xb3\\x27\\x76\\xf2\\x6e\\xd3\\x87\\x49\\x88\\xd0\\x3e\\xff\\x5d\\xa9\\x93\\x8c\\xa3'))
        10

        >>> Ledger.get_transaction_height(UInt256(bytes(32)))   # transaction doesn't exist
        -1

        :param hash_: hash identifier of the transaction
        :type hash_: UInt256
        :return: height of the transaction
        """
        pass

    @classmethod
    def get_transaction_signers(cls, hash_: UInt256) -> List[Signer]:
        """
        Gets the VM state of a transaction.

        >>> Ledger.get_transaction_signers(UInt256(b'\\x29\\x41\\x06\\xdb\\x4c\\xf3\\x84\\xa7\\x20\\x4d\\xba\\x0a\\x04\\x03\\x72\\xb3\\x27\\x76\\xf2\\x6e\\xd3\\x87\\x49\\x88\\xd0\\x3e\\xff\\x5d\\xa9\\x93\\x8c\\xa3'))
        [
            {
                "account": b'\\xa6\\xea\\xb0\\xae\\xaf\\xb4\\x96\\xa1\\x1b\\xb0|\\x88\\x17\\xcar\\xa5J\\x00\\x12\\x04',
                "scopes": 1,
                "allowed_contracts": [],
                "allowed_groups": [],
                "rules": [],
            },
        ]

        :param hash_: hash identifier of the transaction
        :type hash_: UInt256
        :return: VM state of the transaction
        """
        pass

    @classmethod
    def get_transaction_vm_state(cls, hash_: UInt256) -> VMState:
        """
        Gets the VM state of a transaction.

        >>> Ledger.get_transaction_vm_state(UInt256(b'\\x29\\x41\\x06\\xdb\\x4c\\xf3\\x84\\xa7\\x20\\x4d\\xba\\x0a\\x04\\x03\\x72\\xb3\\x27\\x76\\xf2\\x6e\\xd3\\x87\\x49\\x88\\xd0\\x3e\\xff\\x5d\\xa9\\x93\\x8c\\xa3'))
        VMState.HALT

        :param hash_: hash identifier of the transaction
        :type hash_: UInt256
        :return: VM state of the transaction
        """
        pass



from typing import Any, Dict, Optional

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.type.primitive.ibytestringtype import IByteStringType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class LowerMethod(IBuiltinMethod):
    def __init__(self, self_type: IByteStringType = None):
        if not isinstance(self_type, IByteStringType):
            from boa3.internal.model.type.type import Type
            self_type = Type.bytes

        identifier = 'lower'
        args: Dict[str, Variable] = {'self': Variable(self_type)}

        super().__init__(identifier, args, return_type=self_type)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.neo.vm.type.Integer import Integer

        upper_a = Integer(ord('A')).to_byte_array()
        upper_z = Integer(ord('Z')).to_byte_array()

        # string = arg
        code_generator.duplicate_stack_top_item()
        # max_index = len(arg)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)

        # index = 0
        code_generator.convert_literal(0)

        # while index < max_index
        while_begin = code_generator.convert_begin_while()

        #   substr_left = string[:index]
        code_generator.swap_reverse_stack_items(3)
        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_item(4)
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_get_sequence_beginning()
        code_generator.swap_reverse_stack_items(3, rotate=True)
        code_generator.swap_reverse_stack_items(3, rotate=True)

        #   modifier = 1, since using upper is only supported with ASCII for now
        code_generator.duplicate_stack_item(2)  # TODO: verify if string[index] < c0 when other values are implemented
        code_generator.duplicate_stack_item(2)
        code_generator.convert_literal(1)

        #   substr_middle = string[index:index+modifier]
        code_generator.convert_get_substring(is_internal=True)

        #   if 'A' <= substr_middle <= 'Z':
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(upper_a)
        if_lower_than_a = code_generator.convert_begin_if()
        code_generator.change_jump(if_lower_than_a, Opcode.JMPLT)

        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(upper_z)
        if_greater_than_z = code_generator.convert_begin_if()
        code_generator.change_jump(if_greater_than_z, Opcode.JMPGT)

        #       substr_middle = lower(substr_middle)
        code_generator.convert_literal(ord('a') - ord('A'))
        code_generator.convert_operation(BinaryOp.Add, is_internal=True)
        code_generator.convert_cast(self.type)

        code_generator.convert_end_if(if_greater_than_z)
        code_generator.convert_end_if(if_lower_than_a)

        #   substr_right = string[index+modifier:]
        code_generator.swap_reverse_stack_items(3, rotate=True)
        code_generator.swap_reverse_stack_items(3, rotate=True)
        code_generator.insert_opcode(Opcode.INC)
        code_generator.convert_get_sequence_ending()

        concat = BinaryOp.Concat.build(self.type, self.type)  # ensure the correct type on generator stack
        #   string = substr_left + substr_middle + substr_right
        code_generator.convert_operation(concat, is_internal=True)
        code_generator.convert_operation(concat)
        code_generator.remove_stack_item(2)
        code_generator.swap_reverse_stack_items(3)

        #   index += 1
        code_generator.insert_opcode(Opcode.INC)

        while_condition = code_generator.bytecode_size
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_operation(BinaryOp.Gt, is_internal=True)

        code_generator.convert_end_while(while_begin, while_condition, is_internal=True)

        # clean stack
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, IByteStringType):
            return LowerMethod(value)
        return super().build(value)



import ast
from typing import Dict

from boa3.internal.model import set_internal_call
from boa3.internal.model.builtin.interop.interopmethod import InteropMethod
from boa3.internal.model.builtin.interop.runtime.notificationtype import NotificationType
from boa3.internal.model.variable import Variable


class GetNotificationsMethod(InteropMethod):

    def __init__(self, notification_type: NotificationType):
        from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type
        from boa3.internal.model.type.type import Type

        identifier = 'get_notifications'
        syscall = 'System.Runtime.GetNotifications'
        uint160 = UInt160Type.build()

        args: Dict[str, Variable] = {'script_hash': Variable(Type.optional.build(uint160))}
        args_default = set_internal_call(ast.parse("{0}".format(Type.none.default_value)
                                                   ).body[0].value)

        super().__init__(identifier, syscall, args, [args_default],
                         return_type=Type.list.build([notification_type]))

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.model.type.type import Type

        # if arg is not None:
        code_generator.duplicate_stack_top_item()
        arg_is_null = code_generator.convert_begin_if()

        #   if arg == 0:
        code_generator.duplicate_stack_top_item()
        code_generator.convert_cast(Type.int, is_internal=True)
        code_generator.convert_literal(0)
        code_generator.convert_operation(BinaryOp.NumEq, is_internal=True)

        arg_is_zero = code_generator.convert_begin_if()
        #       arg = None
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(None)
        code_generator.convert_end_if(arg_is_zero, is_internal=True)

        code_generator.convert_end_if(arg_is_null, is_internal=True)
        super().generate_internal_opcodes(code_generator)



from typing import Dict, Any, Optional

from boa3.internal.neo3.core.types import UInt160
from boa3_test.test_drive.model.wallet import utils
from boa3_test.test_drive.model.wallet.account import Account


class NeoExpressAccount(Account):
    def __init__(self, script_hash: UInt160,
                 account_version: int = 53,  # default Neo account version
                 name: str = None,
                 label: str = None):

        super().__init__(script_hash, account_version, name=name, label=label)

    @property
    def address(self) -> Optional[str]:
        return utils.address_from_script_hash(self._script_hash.to_array(), self._version)

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> Account:
        address = json['script-hash']
        label = json['label']
        account = NeoExpressAccount(script_hash=UInt160(utils.address_to_script_hash(address)),
                                    label=label)
        return account



from __future__ import annotations

from typing import Optional, List, Dict, Any

from boa3.internal.neo import from_hex_str
from boa3.internal.neo3.core.types import UInt256
from boa3_test.test_drive.model.network.payloads.testtransaction import TestTransaction


class TestBlock:
    def __init__(self):
        self._index: int = 0
        self._timestamp: int = 0
        self._hash: Optional[UInt256] = None
        self._transactions: List[TestTransaction] = []

    @property
    def index(self) -> int:
        return self._index

    @property
    def timestamp(self) -> int:
        return self._timestamp

    def to_json(self) -> Dict[str, Any]:
        json_block = {
            'index': self._index,
            'timestamp': self._timestamp,
            'transactions': [tx.to_json() for tx in self._transactions]
        }

        if self._hash is not None:
            json_block['hash'] = str(self._hash)

        return json_block

    @classmethod
    def from_json(cls, json: Dict[str, Any], *args, **kwargs) -> TestBlock:
        block = object.__new__(cls)

        if 'index' in json:
            block._index = int(json['index'])
        else:
            block._index = 0

        if 'timestamp' in json:
            block._timestamp = int(json['timestamp'])
        else:
            block._timestamp = 0

        if 'transactions' in json:
            tx_json = json['transactions']
            if not isinstance(tx_json, list):
                tx_json = [tx_json]
        else:
            tx_json = []
        block._transactions = [TestTransaction.from_json(js) for js in tx_json]

        if 'hash' in json and isinstance(json['hash'], str):
            block._hash = UInt256(from_hex_str(json['hash']))
        else:
            block._hash = None

        return block

    def __repr__(self) -> str:
        return str(self._hash) if self._hash is not None else str(self._index)



from boa3.internal.model.builtin.decorator.builtindecorator import IBuiltinDecorator


class MetadataDecorator(IBuiltinDecorator):
    def __init__(self):
        identifier = 'metadata'
        super().__init__(identifier)



from __future__ import annotations

from typing import Any, Dict, List, Optional

from boa3_test.test_drive.model.network.payloads.testblock import TestBlock
from boa3_test.tests.test_classes.transaction import Transaction


class Block(TestBlock):
    def __init__(self, index: int):
        self._transactions: List[Transaction] = []
        super().__init__()

        import time
        # time() returns timestamp in nanoseconds and Neo uses timestamp in milliseconds
        self._timestamp: int = int(time.time_ns() / 1_000_000)
        self._index: int = index

    def get_transactions(self) -> List[Transaction]:
        """
        Gets a list of the block transactions. Changes in those transactions don't affect the ones inside the block.

        :return: the block transactions
        :rtype: List[Transaction]
        """
        return [tx.copy() for tx in self._transactions]

    def add_transaction(self, tx: Transaction):
        if all(block_tx != tx for block_tx in self._transactions):
            self._transactions.append(tx)

    @property
    def hash(self) -> Optional[bytes]:
        if self._hash is None:
            return None
        else:
            return self._hash.to_array()

    def to_json(self) -> Dict[str, Any]:
        json_block = super().to_json()

        if self._hash is not None:
            json_block['hash'] = str(self._hash)

        return json_block

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> Block:
        block: Block = super().from_json(json)

        # 'index' and 'timestamp' fields are required
        block._index = int(json['index'])
        block._timestamp = int(json['timestamp'])

        return block



__all__ = [
    'FindOptions',
    'StorageContext',
    'StorageMap',
    'get',
    'get_context',
    'get_read_only_context',
    'put',
    'delete',
    'find',
]


from typing import Union

from boa3.builtin.interop.iterator import Iterator
from boa3.builtin.interop.storage.findoptions import FindOptions
from boa3.builtin.interop.storage.storagecontext import StorageContext
from boa3.builtin.interop.storage.storagemap import StorageMap


def get_context() -> StorageContext:
    """
    Gets current storage context.

    >>> get_context()       # StorageContext cannot be read outside the blockchain
    _InteropInterface

    :return: the current storage context
    :rtype: StorageContext
    """
    pass


def get(key: bytes, context: StorageContext = get_context()) -> bytes:
    """
    Gets a value from the persistent store based on the given key.

    >>> put(b'unit', 'test')
    ... get(b'unit')
    'test'

    >>> get(b'fake_key')
    ''

    :param key: value identifier in the store
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :return: the value corresponding to given key for current storage context
    :rtype: bytes
    """
    pass


def get_read_only_context() -> StorageContext:
    """
    Gets current read only storage context.

    >>> get_context()       # StorageContext cannot be read outside the blockchain
    _InteropInterface

    :return: the current read only storage context
    :rtype: StorageContext
    """
    pass


def put(key: bytes, value: Union[int, bytes, str], context: StorageContext = get_context()):
    """
    Inserts a given value in the key-value format into the persistent storage.

    >>> put(b'unit', 'test')
    None

    :param key: the identifier in the store for the new value
    :type key: bytes
    :param value: value to be stored
    :type value: int or str or bytes
    :param context: storage context to be used
    :type context: StorageContext
    """
    pass


def delete(key: bytes, context: StorageContext = get_context()):
    """
    Removes a given key from the persistent storage if exists.

    >>> put(b'unit', 'test')
    ... delete()
    ... get(b'unit')
    ''

    :param key: the identifier in the store for the new value
    :type key: bytes
    :param context: storage context to be used
    :type context: StorageContext
    """
    pass


def find(prefix: bytes,
         context: StorageContext = get_context(),
         options: FindOptions = FindOptions.NONE) -> Iterator:
    """
    Searches in the storage for keys that start with the given prefix.

    >>> put(b'a1', 'one')
    ... put(b'a2', 'two')
    ... put(b'a3', 'three')
    ... put(b'b4', 'four')
    ... findIterator = find(b'a')
    ... findResults = []
    ... while findIterator.next():
    ...     findResults.append(findIterator.value)
    ... findResults
    ['one', 'two', 'three']

    :param prefix: prefix to find the storage keys
    :type prefix: bytes
    :param context: storage context to be used
    :type context: StorageContext
    :param options: the options of the search
    :type options: FindOptions
    :return: an iterator with the search results
    :rtype: Iterator
    """
    pass



__all__ = ['IteratorMethod',
           'IteratorNextMethod',
           'IteratorType',
           'IteratorValueMethod']

from boa3.internal.model.builtin.interop.iterator.getiteratorvalue import GetIteratorValue as IteratorValueMethod
from boa3.internal.model.builtin.interop.iterator.iteratorinitmethod import IteratorMethod
from boa3.internal.model.builtin.interop.iterator.iteratornextmethod import IteratorNextMethod
from boa3.internal.model.builtin.interop.iterator.iteratortype import IteratorType



from typing import Any, Dict, Iterable, Optional, Union

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.type.collection.mapping.mutable.dicttype import DictType
from boa3.internal.model.type.collection.sequence.sequencetype import SequenceType
from boa3.internal.model.type.primitive.ibytestringtype import IByteStringType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class JoinMethod(IBuiltinMethod):
    def __init__(self, self_type: IByteStringType = None, iterable_type: Union[SequenceType, DictType] = None):
        from boa3.internal.model.type.type import Type

        if not isinstance(self_type, IByteStringType):
            self_type = Type.bytes

        if not isinstance(iterable_type, (SequenceType, DictType)):
            iterable_type = Type.sequence.build_collection([self_type])

        identifier = 'join'
        args: Dict[str, Variable] = {
            'self': Variable(self_type),
            'iterable': Variable(iterable_type),
        }

        super().__init__(identifier, args, return_type=self_type)

    @property
    def identifier(self) -> str:
        from boa3.internal.model.type.type import Type

        if Type.dict.is_type_of(self._arg_iterable.type):
            return '-{0}_{1}'.format(self._identifier, Type.dict.identifier)

        return self._identifier  # JoinMethod default value for self

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    @property
    def _arg_iterable(self) -> Variable:
        return self.args['iterable']

    def push_self_first(self) -> bool:
        return self.has_self_argument

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp

        code_generator.duplicate_stack_top_item()
        # iterable_size = len(iterable)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        # index = 0
        code_generator.convert_literal(0)
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(2)

        # if iterable_size >= 0:
        if_iterable_not_empty = code_generator.convert_begin_if()
        code_generator.change_jump(if_iterable_not_empty, Opcode.JMPLE)

        from boa3.internal.model.type.type import Type
        if Type.dict.is_type_of(self._arg_iterable.type):
            code_generator.swap_reverse_stack_items(3)
            code_generator.convert_builtin_method_call(Builtin.DictKeys)
            code_generator.swap_reverse_stack_items(3)

        #   joined = iterable[0]
        code_generator.duplicate_stack_item(3)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_get_item(index_inserted_internally=True)
        code_generator.swap_reverse_stack_items(2)
        #   index += 1
        code_generator.insert_opcode(Opcode.INC)
        code_generator.swap_reverse_stack_items(2)

        #   while index < iterable_size:
        while_start = code_generator.convert_begin_while()

        code_generator.duplicate_stack_item(5)
        #       joined = joined + string
        code_generator.convert_operation(BinaryOp.Concat, is_internal=True)
        code_generator.duplicate_stack_item(4)
        code_generator.duplicate_stack_item(3)
        code_generator.convert_get_item(index_inserted_internally=True)
        #       joined = joined + iterable[index]
        code_generator.convert_operation(BinaryOp.Concat, is_internal=True)
        code_generator.swap_reverse_stack_items(2)
        #       index += 1
        code_generator.insert_opcode(Opcode.INC)
        code_generator.swap_reverse_stack_items(2)

        # # while condition and end
        while_condition = code_generator.bytecode_size
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(4)
        code_generator.convert_operation(BinaryOp.Lt)
        code_generator.convert_end_while(while_start, while_condition, is_internal=True)

        #   return joined
        code_generator.convert_cast(Type.str, is_internal=True)

        # elif iterable_size < 0:
        else_iterable_is_empty = code_generator.convert_begin_else(if_iterable_not_empty, is_internal=True)
        #   return ""
        code_generator.convert_literal("")
        code_generator.convert_end_if(else_iterable_is_empty)

        code_generator.remove_stack_item(2)
        code_generator.remove_stack_item(2)
        code_generator.remove_stack_item(2)
        code_generator.remove_stack_item(2)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if not isinstance(value, Iterable):
            value = [value]
        if isinstance(value, list) and len(value) <= 2:
            self_type = self._arg_self.type
            if len(value) < 2:
                value.append(None)
            if isinstance(value[0], type(self_type)) and isinstance(value[1], SequenceType):
                return self
            else:
                return JoinMethod(value[0], value[1])

        return super().build(value)



from typing import Dict, Optional

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class ExitMethod(IBuiltinMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        identifier = 'exit'
        args: Dict[str, Variable] = {}
        super().__init__(identifier, args, return_type=Type.none)

    def generate_internal_opcodes(self, code_generator):
        code_generator.insert_opcode(Opcode.ABORT)

    @property
    def _args_on_stack(self) -> int:
        return 0

    @property
    def _body(self) -> Optional[str]:
        return None



__all__ = [
    'ContractDeployCommand',
    'ContractInvokeCommand',
    'ContractListCommand',
    'ContractRunCommand',
]

from boa3_test.test_drive.neoxp.command.neoexpresscommand.contract.deploy import ContractDeployCommand
from boa3_test.test_drive.neoxp.command.neoexpresscommand.contract.invoke import ContractInvokeCommand
from boa3_test.test_drive.neoxp.command.neoexpresscommand.contract.list import ContractListCommand
from boa3_test.test_drive.neoxp.command.neoexpresscommand.contract.run import ContractRunCommand



from typing import Dict, Optional

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.variable import Variable


class DecimalCeilingMethod(IBuiltinMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        identifier = 'ceil'
        args: Dict[str, Variable] = {'x': Variable(Type.int),
                                     'decimals': Variable(Type.int)}
        super().__init__(identifier, args, return_type=Type.int)

    @property
    def exception_message(self) -> str:
        return "decimals cannot be negative"

    def generate_opcodes(self, code_generator):
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.model.operation.unaryop import UnaryOp

        # if decimals < 0
        code_generator.duplicate_stack_item(2)
        code_generator.convert_literal(0)
        code_generator.convert_operation(BinaryOp.Lt, is_internal=True)
        if_negative_decimal = code_generator.convert_begin_if()
        #   raise error
        code_generator.convert_literal(self.exception_message)
        code_generator.convert_raise_exception()
        code_generator.convert_end_if(if_negative_decimal, is_internal=True)

        # unit = 10 ** decimal
        code_generator.duplicate_stack_top_item()
        code_generator.swap_reverse_stack_items(3)
        code_generator.convert_literal(10)
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_operation(BinaryOp.Pow, is_internal=True)

        # difference = (-x) % unit
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_operation(UnaryOp.Negative, is_internal=True)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_operation(BinaryOp.Mod, is_internal=True)

        # if difference < 0
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(0)
        code_generator.convert_operation(BinaryOp.Lt, is_internal=True)
        if_negative_mod = code_generator.convert_begin_if()
        #   difference = unit + difference
        code_generator.convert_operation(BinaryOp.Add, is_internal=True)
        # else
        else_negative_mod = code_generator.convert_begin_else(if_negative_mod, is_internal=True)
        #   # just clear stack
        code_generator.remove_stack_item(2)
        code_generator.convert_end_if(else_negative_mod, is_internal=True)

        # result = x + difference
        code_generator.convert_operation(BinaryOp.Add, is_internal=True)
        super().generate_opcodes(code_generator)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return



"""
Use Upper Kebab Case to name the standards in this dictionary
"""
from boa3.internal.model.standards.nep11divisiblestandard import Nep11DivisibleStandard
from boa3.internal.model.standards.nep11nondivisiblestandard import Nep11NonDivisibleStandard
from boa3.internal.model.standards.nep17standard import Nep17Standard

neo_standards = {
    'NEP-11': [Nep11NonDivisibleStandard(), Nep11DivisibleStandard()],
    'NEP-17': [Nep17Standard()]
}



import abc
from typing import Dict, List


class NeoExpressCommand:
    def __init__(self, command_id: str, args: List[str] = None, options: Dict[str, str] = None):
        if not isinstance(args, list):
            args = []

        if not isinstance(options, dict):
            options = self._get_options()

        self._command_id = command_id
        self._args = args
        self._options = options

    @abc.abstractmethod
    def _get_options(self) -> Dict[str, str]:
        import os.path
        options = {}

        if hasattr(self, 'input') and isinstance(self.input, str) and os.path.isfile(self.input):
            options['--input'] = self.input

        return options

    def cli_command(self) -> str:
        command = [self._command_id]

        for o in self._options.items():
            command.extend(o)

        for arg in self._args:
            command.append(arg)

        return ' '.join(command)

    def is_command(self, command: str) -> bool:
        if not isinstance(command, str) or len(command) == 0:
            return True

        sep = ' '
        args = command.split(sep)
        command_id_split_size = len(self._command_id.split(sep))
        command_id = sep.join(args[:command_id_split_size])
        if command_id != self._command_id:
            return False

        args = args[command_id_split_size:]
        args_compare_size = len(args)
        if args_compare_size == 0:
            return True
        if args_compare_size > len(self._args):
            return False
        return self._args[:args_compare_size] == args

    def __str__(self) -> str:
        return self._command_id



import ast
from typing import Dict

from boa3.internal.analyser.astanalyser import IAstAnalyser
from boa3.internal.model import set_internal_call
from boa3.internal.model.symbol import ISymbol


class ConstructAnalyser(IAstAnalyser, ast.NodeTransformer):
    """
    This class is responsible for pre processing Python constructs

    The methods with the name starting with 'visit_' are implementations of methods from the :class:`NodeVisitor` class.
    These methods are used to walk through the Python abstract syntax tree.
    """

    def __init__(self, analyser, ast_tree: ast.AST, symbol_table: Dict[str, ISymbol],
                 log: bool = False, fail_fast: bool = True):
        super().__init__(ast_tree, root_folder=analyser.root, log=log, fail_fast=fail_fast)
        self.symbols = symbol_table.copy()
        self.analyse_visit(self._tree)

    @property
    def tree(self) -> ast.AST:
        """
        Gets the analysed abstract syntax tree

        :return: the analysed ast
        """
        return self._tree

    def visit_Call(self, call: ast.Call) -> ast.AST:
        """
        Visitor of a function call node

        :param call: the python ast function call node
        """
        if isinstance(call.func, ast.Name):
            from boa3.internal.model.builtin.method import ScriptHashMethod
            to_script_hash = None
            for symbol_id, symbol in self.symbols.items():
                if isinstance(symbol, ScriptHashMethod) and call.func.id == symbol_id:
                    to_script_hash = symbol
                    break

            if to_script_hash is not None:
                from boa3.internal.model.type.type import Type
                types = {
                    Type.int.identifier: int,
                    Type.str.identifier: str,
                    Type.bytes.identifier: bytes
                }

                if len(call.args) != 1:
                    return call

                if isinstance(call.args[0], ast.Constant):
                    value = ast.literal_eval(call.args[0])
                    if not isinstance(value, tuple(types.values())):
                        return call
                else:
                    return call

                from boa3.internal.neo import to_script_hash
                # value must be bytes
                if isinstance(value, int):
                    from boa3.internal.neo.vm.type.Integer import Integer
                    value = Integer(value).to_byte_array()
                elif isinstance(value, str):
                    from boa3.internal.neo.vm.type.String import String
                    value = String(value).to_bytes()
                return set_internal_call(self.parse_to_node(f"UInt160({str(to_script_hash(value))})", call))

        return call



__all__ = [
    'OracleEnableCommand',
    'OracleListCommand',
    'OracleRequestsCommand',
    'OracleResponseCommand',
]

from boa3_test.test_drive.neoxp.command.neoexpresscommand.oracle.enable import OracleEnableCommand
from boa3_test.test_drive.neoxp.command.neoexpresscommand.oracle.list import OracleListCommand
from boa3_test.test_drive.neoxp.command.neoexpresscommand.oracle.requests import OracleRequestsCommand
from boa3_test.test_drive.neoxp.command.neoexpresscommand.oracle.response import OracleResponseCommand



from boa3.internal.model.builtin.method import ScriptHashMethod
from boa3.internal.model.type.collection.sequence.ecpointtype import ECPointType
from boa3.internal.model.type.type import IType


class ECPointToScriptHashMethod(ScriptHashMethod):

    def __init__(self, data_type: IType = None):
        ec_point = ECPointType.build()
        if not ec_point.is_type_of(data_type):
            data_type = ec_point

        super().__init__(data_type)

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal import constants
        from boa3.internal.model.builtin.interop.crypto import CheckSigMethod
        from boa3.internal.model.builtin.interop.interop import Interop
        from boa3.internal.model.operation.binaryop import BinaryOp

        ecpoint_default = ECPointType.build().default_value

        # build CheckSig script
        from boa3.internal.neo.vm.opcode import OpcodeHelper
        push_ecpoint_opcode, push_ecpoint_data = OpcodeHelper.get_pushdata_and_data_from_size(len(ecpoint_default))
        code_generator.convert_literal(push_ecpoint_opcode + push_ecpoint_data)
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_literal(CheckSigMethod.get_raw_bytes())
        code_generator.convert_operation(BinaryOp.Concat, is_internal=True)
        code_generator.convert_operation(BinaryOp.Concat, is_internal=True)

        # to_script_hash
        code_generator.convert_builtin_method_call(Interop.Sha256, is_internal=True)
        code_generator.convert_builtin_method_call(Interop.Ripemd160, is_internal=True)

        # limit result to UInt160 length
        code_generator.convert_literal(0)
        code_generator.convert_literal(constants.SIZE_OF_INT160)
        code_generator.convert_get_substring(is_internal=True)
        code_generator.convert_cast(self.return_type, is_internal=True)



from typing import Dict

from boa3_test.test_drive.neoxp.command.neoexpresscommand.oracle.ioraclecommand import IOracleCommand


class OracleRequestsCommand(IOracleCommand):
    def __init__(self,
                 neo_express_data_file: str = None):

        self.input = neo_express_data_file

        command_id = 'requests'

        super().__init__(command_id)

    def _get_options(self) -> Dict[str, str]:
        return super()._get_options()



from boa3.internal.model.builtin.method.printmethod import PrintMethod


class PrintIntMethod(PrintMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        arg_value = Type.int

        super().__init__(arg_value)

    def _generate_print_opcodes(self, code_generator):
        from boa3.internal.model.builtin.interop.stdlib.itoamethod import ItoaMethod

        itoa_method = ItoaMethod(internal_call_args=1)
        code_generator.convert_builtin_method_call(itoa_method, is_internal=True)



import ast
from typing import Dict, List, Optional, Tuple

from boa3.internal.model.builtin.classmethod.indexmethod import IndexMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.collection.sequence.sequencetype import SequenceType
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode import OpcodeHelper
from boa3.internal.neo.vm.opcode.Opcode import Opcode
from boa3.internal.neo.vm.type.Integer import Integer
from boa3.internal.neo.vm.type.String import String


class IndexSequenceMethod(IndexMethod):

    def __init__(self, sequence_type: Optional[SequenceType] = None, arg_value: Optional[IType] = None):
        from boa3.internal.model.type.type import Type
        if not isinstance(sequence_type, SequenceType):
            sequence_type = Type.sequence
        if not isinstance(arg_value, IType):
            arg_value = Type.any

        args: Dict[str, Variable] = {
            'self': Variable(sequence_type),
            'x': Variable(arg_value),
            'start': Variable(Type.int),
            'end': Variable(Type.int),
        }

        start_default = ast.parse("{0}".format(0)
                                  ).body[0].value
        end_default = ast.parse("-1").body[0].value.operand
        end_default.n = -1

        super().__init__(args, defaults=[start_default, end_default])

    def validate_parameters(self, *params: IExpression) -> bool:
        if 2 <= len(params) <= 4:
            return False
        if not all(isinstance(param, IExpression) for param in params):
            return False

        from boa3.internal.model.type.itype import IType
        sequence_type: IType = params[0].type

        if not isinstance(sequence_type, SequenceType):
            return False
        from boa3.internal.model.type.collection.sequence.mutable.listtype import ListType
        from boa3.internal.model.type.collection.sequence.tupletype import TupleType
        from boa3.internal.model.type.collection.sequence.rangetype import RangeType
        if not isinstance(sequence_type, (ListType, TupleType, RangeType)):
            return False
        return True

    @property
    def exception_message(self) -> str:
        return "x not in sequence"

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp

        # receives: end, start, x, sequence
        code_generator.swap_reverse_stack_items(4)

        # if index < 0:
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(0)
        is_negative_index = code_generator.convert_begin_if()
        # # if index >= 0, jump to verify_big_end
        code_generator.change_jump(is_negative_index, Opcode.JMPGE)

        #   # gets the correspondent positive value of end
        #   end = end + len(sequence) + 1
        code_generator.duplicate_stack_item(4)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_operation(BinaryOp.Add, is_internal=True)
        code_generator.insert_opcode(Opcode.INC)

        #   if index < 0:
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(0)
        is_still_negative_index = code_generator.convert_begin_if()
        #   # if end is not negative anymore, start verifying start
        code_generator.change_jump(is_still_negative_index, Opcode.JMPGE)

        #       index = 0
        code_generator.remove_stack_top_item()
        code_generator.convert_literal(0)

        code_generator.convert_end_if(is_negative_index, is_internal=True)

        # if end > len(sequence)
        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_item(5)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        is_big_end = code_generator.convert_begin_if()
        # # if end <= len(sequence), start verifying start
        code_generator.change_jump(is_big_end, Opcode.JMPLE)

        #   end = len(sequence)
        code_generator.remove_stack_top_item()
        code_generator.duplicate_stack_item(3)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)

        code_generator.convert_end_if(is_still_negative_index, is_internal=True)
        code_generator.convert_end_if(is_big_end, is_internal=True)
        code_generator.swap_reverse_stack_items(2)

        # if index < 0:
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(0)
        is_negative_index = code_generator.convert_begin_if()
        # # if index >= 0, jump to verify_big_start
        code_generator.change_jump(is_negative_index, Opcode.JMPGE)

        #   # gets the correspondent positive value of end
        #   start = start + len(sequence) + 1
        code_generator.duplicate_stack_item(4)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_operation(BinaryOp.Add, is_internal=True)
        code_generator.insert_opcode(Opcode.INC)

        #   if index < 0:
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(0)
        is_still_negative_index = code_generator.convert_begin_if()
        #   # if start is not negative anymore
        code_generator.change_jump(is_still_negative_index, Opcode.JMPGE)

        #       index = 0
        code_generator.remove_stack_top_item()
        code_generator.convert_literal(0)

        code_generator.convert_end_if(is_negative_index, is_internal=True)

        # if start > len(sequence)
        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_item(5)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        is_big_end = code_generator.convert_begin_if()
        # # if start <= len(sequence), start verifying start
        code_generator.change_jump(is_big_end, Opcode.JMPLE)

        #   start = len(sequence)
        code_generator.remove_stack_top_item()
        code_generator.duplicate_stack_item(3)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)

        code_generator.convert_end_if(is_still_negative_index, is_internal=True)
        code_generator.convert_end_if(is_big_end, is_internal=True)

        # begin while
        # while self[index] != x:
        while_start = code_generator.convert_begin_while()

        #   index += 1
        code_generator.insert_opcode(Opcode.INC)

        while_condition = code_generator.bytecode_size
        #   if end <= index
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(2)
        invalid_index = code_generator.convert_begin_if()
        code_generator.change_jump(invalid_index, Opcode.JMPGT)
        #       raise error
        code_generator.convert_literal(self.exception_message)
        code_generator.convert_raise_exception()

        invalid_index = code_generator.convert_begin_else(invalid_index, is_internal=True)

        code_generator.convert_end_if(invalid_index, is_internal=True)

        #   self[index] == x:
        code_generator.duplicate_stack_item(4)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_get_item(index_inserted_internally=True)
        code_generator.duplicate_stack_item(4)
        code_generator.convert_operation(BinaryOp.NumNotEq, is_internal=True)

        code_generator.convert_end_while(while_start, while_condition, is_internal=True)

        for _ in range(1, len(self.args)):
            code_generator.remove_stack_item(2)

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.internal.compiler.codegenerator import get_bytes_count

        jmp_place_holder = (Opcode.JMP, b'\x01')
        message = String(self.exception_message).to_bytes()

        # receives: end, start, x, sequence
        verify_negative_index = [           # verifies if index in a negative value
            (Opcode.DUP, b''),
            (Opcode.PUSHM1, b''),
            jmp_place_holder                # if index >= 0, jump to verify_big_end or verify_big_start
        ]

        fix_negative_end = [                # gets the correspondent positive value of end
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            (Opcode.ADD, b''),
            (Opcode.INC, b''),              # end = end + len(sequence) + 1
            (Opcode.DUP, b''),
            (Opcode.PUSHM1, b''),
            jmp_place_holder                # if end is not negative anymore, start verifying start
        ]

        fix_still_negative_index = [        # if index is still negative, consider it 0 then
            (Opcode.DROP, b''),
            (Opcode.PUSH0, b''),            # end = 0
        ]

        jmp_fix_negative_index = OpcodeHelper.get_jump_and_data(Opcode.JMPGT, get_bytes_count(fix_negative_end +
                                                                                              fix_still_negative_index), True)
        verify_negative_index[-1] = jmp_fix_negative_index

        verify_big_end = [                  # verify if end is bigger then len(sequence)
            (Opcode.DUP, b''),
            (Opcode.PUSH4, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),
            jmp_place_holder                # if end <= len(sequence), start verifying start
        ]

        fix_big_end = [                     # consider end as len(sequence)
            (Opcode.DROP, b''),
            (Opcode.PUSH2, b''),
            (Opcode.PICK, b''),
            (Opcode.SIZE, b''),             # end = len(sequence)
        ]

        jmp_other_verifies = OpcodeHelper.get_jump_and_data(Opcode.JMPGT, get_bytes_count(fix_still_negative_index +
                                                                                          verify_big_end +
                                                                                          fix_big_end), True)
        fix_negative_end[-1] = jmp_other_verifies

        jmp_fix_big_index = OpcodeHelper.get_jump_and_data(Opcode.JMPLE, get_bytes_count(fix_big_end), True)
        verify_big_end[-1] = jmp_fix_big_index

        verify_and_fix_end = [              # collection of Opcodes regarding verifying and fixing end index
            (Opcode.REVERSE4, b''),
        ]
        verify_and_fix_end.extend(verify_negative_index)
        verify_and_fix_end.extend(fix_negative_end)
        verify_and_fix_end.extend(fix_still_negative_index)
        verify_and_fix_end.extend(verify_big_end)
        verify_and_fix_end.extend(fix_big_end)

        verify_and_fix_start = [            # collection of Opcodes regarding verifying and fixing start index
            (Opcode.SWAP, b'')
        ]
        verify_and_fix_start.extend(verify_negative_index)
        verify_and_fix_start.extend(fix_negative_end)
        verify_and_fix_start.extend(fix_still_negative_index)
        verify_and_fix_start.extend(verify_big_end)
        verify_and_fix_start.extend(fix_big_end)

        verify_while = [                    # verify if already went through all items on the sequence[start:end]
            (Opcode.OVER, b''),             # index = start
            (Opcode.OVER, b''),
            jmp_place_holder                # if index <= start, jump to not_inside_sequence
        ]

        compare_item = [                    # verifies if x is in sequence
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.OVER, b''),
            (Opcode.PICKITEM, b''),
            (Opcode.PUSH3, b''),
            (Opcode.PICK, b''),
            (Opcode.NUMEQUAL, b''),         # found_x = sequence[index] == x
            jmp_place_holder                # if found_x, jump to return_index
        ]

        not_found = [                       # increments index and goes back to verify again
            (Opcode.INC, b''),              # index++
            # jump to verify_while
        ]

        jmp_back_to_verify = OpcodeHelper.get_jump_and_data(Opcode.JMP, -get_bytes_count(verify_while +
                                                                                         compare_item +
                                                                                         not_found), True)
        not_found.append(jmp_back_to_verify)

        jmp_to_error = OpcodeHelper.get_jump_and_data(Opcode.JMPLE, get_bytes_count(compare_item +
                                                                                    not_found), True)
        verify_while[-1] = jmp_to_error

        not_inside_sequence = [             # send error message saying that x is not in sequence
            (Opcode.PUSHDATA1, Integer(len(message)).to_byte_array(signed=True, min_length=1) + message),
            (Opcode.THROW, b''),
        ]

        jmp_to_return_index = OpcodeHelper.get_jump_and_data(Opcode.JMPIF, get_bytes_count(not_found +
                                                                                           not_inside_sequence), True)
        compare_item[-1] = jmp_to_return_index

        return_index = [                    # removes all values in the stack but the index
            (Opcode.NIP, b''),
            (Opcode.NIP, b''),
            (Opcode.NIP, b''),
        ]

        return (
            verify_and_fix_end +
            verify_and_fix_start +
            verify_while +
            compare_item +
            not_found +
            not_inside_sequence +
            return_index
        )



from abc import ABC
from typing import Any, Dict, Optional, Sequence

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.identifiedsymbol import IdentifiedSymbol
from boa3.internal.model.type.itype import IType
from boa3.internal.model.type.primitive.bytestype import BytesType
from boa3.internal.model.type.primitive.inttype import IntType
from boa3.internal.model.type.primitive.strtype import StrType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class ToBytesMethod(IBuiltinMethod, ABC):
    def __init__(self, self_type: IType):
        identifier = 'to_bytes'
        args: Dict[str, Variable] = {'self': Variable(self_type)}
        from boa3.internal.model.type.type import Type
        super().__init__(identifier, args, return_type=Type.bytes)

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1:
            return False
        return isinstance(params[0], IExpression) and isinstance(params[0].type, BytesType)

    @property
    def identifier(self) -> str:
        if isinstance(self._arg_self.type, IdentifiedSymbol):
            return '-{0}_{1}'.format(self._arg_self.type.identifier, self._identifier)
        return self._identifier

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.type.type import Type
        code_generator.convert_cast(Type.bytes, is_internal=True)

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, IntType):
            return IntToBytesMethod(value)
        elif isinstance(value, StrType):
            return StrToBytesMethod(value)
        # if it is not a valid type, show mismatched type with int
        return IntToBytesMethod()

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None


class _ConvertToBytesMethod(ToBytesMethod):
    def __init__(self):
        super().__init__(None)


ToBytes = _ConvertToBytesMethod()


class IntToBytesMethod(ToBytesMethod):
    def __init__(self, self_type: IType = None):
        if not isinstance(self_type, IntType):
            from boa3.internal.model.type.type import Type
            self_type = Type.int
        super().__init__(self_type)

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, Sequence) and len(value) == 1:
            value = value[0]
        if type(value) == type(self.args['self'].type):
            return self
        if isinstance(value, IntType):
            return IntToBytesMethod(value)
        return super().build(value)

    def generate_internal_opcodes(self, code_generator):
        code_generator.duplicate_stack_top_item()
        code_generator.insert_opcode(Opcode.NZ)
        is_zero = code_generator.convert_begin_if()
        # if number != 0:
        #   generic implementation
        super().generate_internal_opcodes(code_generator)
        # else:
        is_zero = code_generator.convert_begin_else(is_zero, is_internal=True)
        #   result = b'\x00'
        code_generator.remove_stack_top_item()
        code_generator.convert_literal(b'\x00')

        code_generator.convert_end_if(is_zero, is_internal=True)


class StrToBytesMethod(ToBytesMethod):
    def __init__(self, self_type: IType = None):
        if not isinstance(self_type, StrType):
            from boa3.internal.model.type.type import Type
            self_type = Type.str
        super().__init__(self_type)

    def generate_internal_opcodes(self, code_generator):
        # string and bytes' stack item are the same
        pass

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, Sequence) and len(value) == 1:
            value = value[0]
        if type(value) == type(self.args['self'].type):
            return self
        if isinstance(value, StrType):
            return StrToBytesMethod(value)
        return super().build(value)



from __future__ import annotations

from typing import Any, Dict, Optional

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.method import Method
from boa3.internal.model.property import Property
from boa3.internal.model.type.classes.classarraytype import ClassArrayType
from boa3.internal.model.variable import Variable


class TransactionType(ClassArrayType):
    """
    A class used to represent Neo Transaction class
    """

    def __init__(self):
        super().__init__('Transaction')
        from boa3.internal.model.type.type import Type
        from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type
        from boa3.internal.model.type.collection.sequence.uint256type import UInt256Type

        self._variables: Dict[str, Variable] = {
            'hash': Variable(UInt256Type.build()),
            'version': Variable(Type.int),
            'nonce': Variable(Type.int),
            'sender': Variable(UInt160Type.build()),
            'system_fee': Variable(Type.int),
            'network_fee': Variable(Type.int),
            'valid_until_block': Variable(Type.int),
            'script': Variable(Type.bytes),
        }
        self._constructor: Method = None

    @property
    def class_variables(self) -> Dict[str, Variable]:
        return {}

    @property
    def instance_variables(self) -> Dict[str, Variable]:
        return self._variables.copy()

    @property
    def properties(self) -> Dict[str, Property]:
        return {}

    @property
    def static_methods(self) -> Dict[str, Method]:
        return {}

    @property
    def class_methods(self) -> Dict[str, Method]:
        return {}

    @property
    def instance_methods(self) -> Dict[str, Method]:
        return {}

    def constructor_method(self) -> Optional[Method]:
        # was having a problem with recursive import
        if self._constructor is None:
            self._constructor: Method = TransactionMethod(self)
        return self._constructor

    @classmethod
    def build(cls, value: Any = None) -> TransactionType:
        if value is None or cls._is_type_of(value):
            return _Transaction

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, TransactionType)


_Transaction = TransactionType()


class TransactionMethod(IBuiltinMethod):

    def __init__(self, return_type: TransactionType):
        identifier = '-Transaction__init__'
        args: Dict[str, Variable] = {}
        super().__init__(identifier, args, return_type=return_type)

    def validate_parameters(self, *params: IExpression) -> bool:
        return len(params) == 0

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.neo3.core.types import UInt160, UInt256

        uint160_default = UInt160.zero().to_array()
        uint256_default = UInt256.zero().to_array()

        code_generator.convert_literal(b'')  # script
        code_generator.convert_literal(0)  # valid_until_block
        code_generator.convert_literal(0)  # network_fee
        code_generator.convert_literal(0)  # system_fee
        code_generator.convert_literal(uint160_default)  # sender
        code_generator.convert_literal(0)  # nonce
        code_generator.convert_literal(0)  # version
        code_generator.convert_literal(uint256_default)  # hash
        code_generator.convert_new_array(length=8, array_type=self.type)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return



__all__ = [
    'ContractManagement',
    'Contract',
]


from typing import Any, Optional

from boa3.builtin.interop.contract import Contract
from boa3.builtin.type import UInt160


class ContractManagement:
    """
    A class used to represent the ContractManagement native contract.

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/reference/scapi/framework/native/ContractManagement>`__
    to learn more about the ContractManagement class.
    """

    hash: UInt160

    @classmethod
    def get_minimum_deployment_fee(cls) -> int:
        """
        Gets the minimum fee of contract deployment.

        >>> ContractManagement.get_minimum_deployment_fee()
        1000000000

        :return: the minimum fee of contract deployment
        """
        pass

    @classmethod
    def get_contract(cls, script_hash: UInt160) -> Optional[Contract]:
        """
        Gets a contract with a given hash. If the script hash is not associated with a smart contract, then it will
        return None.

        >>> ContractManagement.get_contract(UInt160(b'\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2'))    # GAS script hash
        {
            'id': -6,
            'update_counter': 0,
            'hash': b'\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2',
            'nef': b'NEF3neo-core-v3.0\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00#\\x10A\\x1a\\xf7{g@\\x10A\\x1a\\xf7{g@\\x10A\\x1a\\xf7{g@\\x10A\\x1a\\xf7{g@\\x10A\\x1a\\xf7{g@QA\\xc7\\x9e',
            'manifest': {
                'name': 'GasToken',
                'group': [],
                'supported_standards': ['NEP-17'],
                'abi': [[['balanceOf', [['account', 20]], 17, 0, True], ['decimals', [], 17, 7, True], ['symbol', [], 19, 14, True], ['totalSupply', [], 17, 21, True], ['transfer', [['from', 20], ['to', 20], ['amount', 17], ['data', 0]], 16, 28, False]], [['Transfer', [['from', 20], ['to', 20], ['amount', 17]]]]],
                'permissions': [[None, None]],
                'trusts': [],
                'extras': 'null'
            },
        }

        >>> ContractManagement.get_contract(UInt160(bytes(20)))    # there is no smart contract associated with this script hash
        None

        :param script_hash: a smart contract hash
        :type script_hash: UInt160
        :return: a contract
        :rtype: Contract

        :raise Exception: raised if hash length isn't 20 bytes.
        """
        pass

    @classmethod
    def has_method(cls, hash: UInt160, method: str, parameter_count: int) -> bool:
        """
        Check if a method exists in a contract.

        >>> ContractManagement.has_method(UInt160(b'\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2'),
        ...                               'balanceOf', 1)    # GAS script hash
        True

        >>> ContractManagement.has_method(UInt160(b'\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2'),
        ...                               'balanceOf', 10)    # GAS script hash
        False

        >>> ContractManagement.has_method(UInt160(b'\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2'),
        ...                               'invalid', 1)    # GAS script hash
        False

        :param hash: The hash of the deployed contract
        :type hash: UInt160
        :param method: The name of the method
        :type method: str
        :param parameter_count: The number of parameters
        :type parameter_count: int
        :return: whether the method exists or not
        :rtype: bool

        :raise Exception: raised if hash length isn't 20 bytes or if the parameter_count is less than 0.
        """
        pass

    @classmethod
    def deploy(cls, nef_file: bytes, manifest: bytes, data: Any = None) -> Contract:
        """
        Creates a smart contract given the script and the manifest.

        >>> nef_file_ = get_script(); manifest_ = get_manifest()    # get the script and manifest somehow
        ... ContractManagement.deploy(nef_file_, manifest_, None)             # smart contract will be deployed
        {
            'id': 2,
            'update_counter': 0,
            'hash': b'\\x92\\x8f+1q\\x86z_@\\x94\\xf5pE\\xcb\\xb8 \\x0f\\\\`Z',
            'nef': b'NEF3neo3-boa by COZ-1.0.0\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x07W\\x00\\x02xy\\x9e@\\xf9\\7b\\xbb\\xcc',
            'manifest': {
                'name': 'TestContract',
                'group': [],
                'supported_standards': [],
                'abi': [[['test', [['a', 17], ['b', 17]], 17, 0, False]], []],
                'permissions': [],
                'trusts': [],
                'extras': 'null'
            },
        }

        :param nef_file: the target smart contract's compiled nef
        :type nef_file: bytes
        :param manifest: the manifest.json that describes how the script should behave
        :type manifest: bytes
        :param data: the parameters for the _deploy function
        :type data: Any

        :return: the contract that was created
        :rtype: Contract

        :raise Exception: raised if the nef or the manifest are not a valid smart contract.
        """
        pass

    @classmethod
    def update(cls, nef_file: bytes, manifest: bytes, data: Any = None):
        """
        Updates the executing smart contract given the script and the manifest.

        >>> nef_file_ = get_script(); manifest_ = get_manifest()    # get the script and manifest somehow
        ... ContractManagement.update(nef_file_, manifest_, None)             # smart contract will be updated
        None

        :param nef_file: the new smart contract's compiled nef
        :type nef_file: bytes
        :param manifest: the new smart contract's manifest
        :type manifest: bytes
        :param data: the parameters for the _deploy function
        :type data: Any

        :raise Exception: raised if the nef and the manifest are not a valid smart contract or the new contract is the
            same as the old one.
        """
        pass

    @classmethod
    def destroy(cls):
        """
        Destroy the executing smart contract.

        >>> ContractManagement.destroy()
        None

        """
        pass



__all__ = [
    'base58_encode',
    'base58_decode',
    'base58_check_encode',
    'base58_check_decode',
    'base64_encode',
    'base64_decode',
    'serialize',
    'deserialize',
    'atoi',
    'itoa',
    'memory_search',
    'memory_compare',
]


from typing import Any, Union


def base58_encode(key: bytes) -> str:
    """
    Encodes a bytes value using base58.

    >>> base58_encode(b'unit test')
    b"2VhL46g69A1mu"

    :param key: bytes value to be encoded
    :type key: bytes
    :return: the encoded string
    :rtype: str
    """
    pass


def base58_decode(key: str) -> bytes:
    """
    Decodes a string value encoded with base58.

    >>> base58_decode('2VhL46g69A1mu')
    b"unit test"

    :param key: string value to be decoded
    :type key: str
    :return: the decoded bytes
    :rtype: bytes
    """
    pass


def base58_check_encode(key: bytes) -> str:
    """
    Converts a bytes value to its equivalent str representation that is encoded with base-58 digits. The encoded str
    contains the checksum of the binary data.

    >>> base58_check_encode(b'unit test')
    b"AnJcKqvgBwKxsjX75o"

    :param key: bytes value to be encoded
    :type key: bytes
    :return: the encoded string
    :rtype: str
    """
    pass


def base58_check_decode(key: str) -> bytes:
    """
    Converts the specified str, which encodes binary data as base-58 digits, to an equivalent bytes value. The encoded
    str contains the checksum of the binary data.

    >>> base58_check_decode('AnJcKqvgBwKxsjX75o')
    b"unit test"

    :param key: string value to be decoded
    :type key: str
    :return: the decoded bytes
    :rtype: bytes
    """
    pass


def base64_encode(key: bytes) -> str:
    """
    Encodes a bytes value using base64.

    >>> base64_encode(b'unit test')
    b"dW5pdCB0ZXN0"

    :param key: bytes value to be encoded
    :type key: bytes
    :return: the encoded string
    :rtype: str
    """
    pass


def base64_decode(key: str) -> bytes:
    """
    Decodes a string value encoded with base64.

    >>> base64_decode("dW5pdCB0ZXN0")
    b"unit test"

    :param key: string value to be decoded
    :type key: str
    :return: the decoded bytes
    :rtype: bytes
    """
    pass


def serialize(item: Any) -> bytes:
    """
    Serializes the given value into its bytes representation.

    >>> serialize('42')
    b'(\x0242'

    >>> serialize(42)
    b'!\x01*'

    >>> serialize([2, 3, 5, 7])
    b'@\x04!\x01\x02!\x01\x03!\x01\x05!\x01\x07'

    >>> serialize({1: 1, 2: 1, 3: 2})
    b'H\x03!\x01\x01!\x01\x01!\x01\x02!\x01\x01!\x01\x03!\x01\x02'

    :param item: value to be serialized
    :type item: Any
    :return: the serialized value
    :rtype: bytes

    :raise Exception: raised if the item's type is not serializable.
    """
    pass


def deserialize(data: bytes) -> Any:
    """
    Deserializes the given bytes value.

    >>> deserialize(b'(\x0242')
    '42'

    >>> deserialize(b'!\x01*')
    42

    >>> deserialize(b'@\x04!\x01\x02!\x01\x03!\x01\x05!\x01\x07')
    [2, 3, 5, 7]

    >>> deserialize(b'H\x03!\x01\x01!\x01\x01!\x01\x02!\x01\x01!\x01\x03!\x01\x02')
    {1: 1, 2: 1, 3: 2}

    :param data: serialized value
    :type data: bytes
    :return: the deserialized result
    :rtype: Any

    :raise Exception: raised when the date doesn't represent a serialized value.
    """
    pass


def atoi(value: str, base: int = 10) -> int:
    """
    Converts a character string to a specific base value, decimal or hexadecimal. The default is decimal.

    >>> atoi('10')
    10

    >>> atoi('123')
    123

    >>> atoi('1f', 16)
    31

    >>> atoi('ff', 16)
    -1

    :param value: the int value as a string
    :type value: str
    :param base: the value base
    :type base: int
    :return: the equivalent value
    :rtype: int

    :raise Exception: raised when base isn't 10 or 16.
    """
    pass


def itoa(value: int, base: int = 10) -> str:
    """
    Converts the specific type of value to a decimal or hexadecimal string. The default is decimal.

    >>> itoa(10)
    '10'

    >>> itoa(123)
    '123'

    >>> itoa(-1, 16)
    'f'

    >>> itoa(15, 16)
    '0f'

    :param value: the int value
    :type value: int
    :param base: the value's base
    :type base: int
    :return: the converted string
    :rtype: int
    """
    pass


def memory_search(mem: Union[bytes, str], value: Union[bytes, str], start: int = 0, backward: bool = False) -> int:
    """
    Searches for a given value in a given memory.

    >>> memory_search('abcde', 'a', 0)
    0

    >>> memory_search('abcde', 'e', 0)
    4

    :param mem: the memory
    :type mem: bytes or str
    :param value: the value
    :type value: bytes or str
    :param start: the index the search should start from
    :type start: int
    :param backward: whether it should invert the memory
    :type backward: bool

    :return: the index of the value in the memory. Returns -1 if it's not found
    :rtype: int
    """
    pass


def memory_compare(mem1: Union[bytes, str], mem2: Union[bytes, str]) -> int:
    """
    Compares a memory with another one.

    >>> memory_compare('abc', 'abc')
    0

    >>> memory_compare('ABC', 'abc')
    -1

    >>> memory_compare('abc', 'ABC')
    1

    :param mem1: a memory to be compared to another one
    :type mem1: bytes or str
    :param mem2: a memory that will be compared with another one
    :type mem2: bytes or str

    :return: -1 if mem1 precedes mem2, 0 if mem1 and mem2 are equal, 1 if mem1 follows mem2
    :rtype: int
    """
    pass



from typing import Dict

from boa3.internal.model.builtin.method.strmethod import StrMethod
from boa3.internal.model.variable import Variable


class StrBoolMethod(StrMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        args: Dict[str, Variable] = {
            'object': Variable(Type.bool),
        }

        super().__init__(args)

    def generate_internal_opcodes(self, code_generator):
        # if object == True:
        if_address = code_generator.convert_begin_if()
        #   result = 'True'
        code_generator.convert_literal(str(True))

        # else:
        if_address = code_generator.convert_begin_else(if_address, is_internal=True)
        #   result = 'False'
        code_generator.convert_literal(str(False))

        code_generator.convert_end_if(if_address, is_internal=True)



from abc import ABC, abstractmethod
from argparse import ArgumentParser, _SubParsersAction


class ICommand(ABC):

    def __init__(self, main_parser: _SubParsersAction, command_name: str, help_text: str):
        self.parser: ArgumentParser = main_parser.add_parser(command_name, help=help_text)

    @abstractmethod
    def add_arguments_and_callback(self):
        pass

    @staticmethod
    @abstractmethod
    def execute_command(args: dict):
        pass



__all__ = [
    'CryptoLib',
    'NamedCurve',
]

from typing import Any

from boa3.builtin.interop.crypto import NamedCurve
from boa3.builtin.type import ECPoint, UInt160


class CryptoLib:
    """
    A class used to represent the CryptoLib native contract.

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/reference/scapi/framework/native/CryptoLib>`__
    to learn more about the CryptoLib class.
    """

    hash: UInt160

    @classmethod
    def murmur32(cls, data: bytes, seed: int) -> bytes:
        """
        Computes the hash value for the specified byte array using the murmur32 algorithm.

        >>> CryptoLib.murmur32(b'unit test', 0)
        b"\\x90D'G"

        :param data: the input to compute the hash code for
        :type data: bytes
        :param seed: the seed of the murmur32 hash function
        :type seed: int
        :return: the hash value
        :rtype: bytes
        """
        pass

    @classmethod
    def sha256(cls, key: Any) -> bytes:
        """
        Encrypts a key using SHA-256.

        >>> CryptoLib.sha256('unit test')
        b'\\xdau1>J\\xc2W\\xf8LN\\xfb2\\x0f\\xbd\\x01\\x1cr@<\\xf5\\x93<\\x90\\xd2\\xe3\\xb8$\\xd6H\\x96\\xf8\\x9a'

        >>> CryptoLib.sha256(10)
        b'\\x9c\\x82r\\x01\\xb9@\\x19\\xb4/\\x85pk\\xc4\\x9cY\\xff\\x84\\xb5`M\\x11\\xca\\xaf\\xb9\\n\\xb9HV\\xc4\\xe1\\xddz'

        :param key: the key to be encrypted
        :type key: Any
        :return: a byte value that represents the encrypted key
        :rtype: bytes
        """
        pass

    @classmethod
    def ripemd160(cls, key: Any) -> bytes:
        """
        Encrypts a key using RIPEMD-160.

        >>> CryptoLib.ripemd160('unit test')
        b'H\\x8e\\xef\\xf4Zh\\x89:\\xe6\\xf1\\xdc\\x08\\xdd\\x8f\\x01\\rD\\n\\xbdH'

        >>> CryptoLib.ripemd160(10)
        b'\\xc0\\xda\\x02P8\\xed\\x83\\xc6\\x87\\xdd\\xc40\\xda\\x98F\\xec\\xb9\\x7f9\\x98'

        :param key: the key to be encrypted
        :type key: Any
        :return: a byte value that represents the encrypted key
        :rtype: bytes
        """
        pass

    @classmethod
    def verify_with_ecdsa(cls, message: bytes, pubkey: ECPoint, signature: bytes, curve: NamedCurve) -> bool:
        """
        Using the elliptic curve, it checks if the signature of the message was originally produced by the public key.

        >>> CryptoLib.verify_with_ecdsa(b'unit test', ECPoint(b'\\x03\\x5a\\x92\\x8f\\x20\\x16\\x39\\x20\\x4e\\x06\\xb4\\x36\\x8b\\x1a\\x93\\x36\\x54\\x62\\xa8\\xeb\\xbf\\xf0\\xb8\\x81\\x81\\x51\\xb7\\x4f\\xaa\\xb3\\xa2\\xb6\\x1a'),
        ...                             b'wrong_signature', NamedCurve.SECP256R1)
        False

        :param message: the encrypted message
        :type message: bytes
        :param pubkey: the public key that might have created the item
        :type pubkey: ECPoint
        :param signature: the signature of the item
        :type signature: bytes
        :param curve: the curve that will be used by the ecdsa
        :type curve: NamedCurve
        :return: a boolean value that represents whether the signature is valid
        :rtype: bool
        """
        pass

    @classmethod
    def bls12_381_add(cls, x: Any, y: Any) -> Any:
        """
        Add operation of two bls12381 points.

        :param x: The first point
        :type x: bytes
        :param y: The second point
        :type y: bytes
        :return: the two points sum
        :rtype: bytes
        """
        pass

    @classmethod
    def bls12_381_deserialize(cls, data: bytes) -> Any:
        """
        Deserialize a bls12381 point.

        :param data: The point as byte array
        :type data: bytes
        :return: the deserialized point
        :rtype: bytes
        """
        pass

    @classmethod
    def bls12_381_equal(cls, x: Any, y: Any) -> bool:
        """
        Determines whether the specified points are equal.

        :param x: The first point
        :type x: bytes
        :param y: The second point
        :type y: bytes
        :return: whether the specified points are equal or not
        :rtype: bool
        """
        pass

    @classmethod
    def bls12_381_mul(cls, x: Any, mul: bytes, neg: bool) -> Any:
        """
        Mul operation of gt point and multiplier.

        :param x: The point
        :type x: bytes
        :param mul: Multiplier, 32 bytes, little-endian
        :type mul: int
        :param neg: negative number
        :type neg: bool
        :return: the two points product
        :rtype: bytes
        """
        pass

    @classmethod
    def bls12_381_pairing(cls, g1: Any, g2: Any) -> Any:
        """
        Pairing operation of g1 and g2.

        :param g1: The g1 point
        :type g1: bytes
        :param g2: The g2 point
        :type g2: bytes
        :return: the two points pairing
        :rtype: bytes
        """
        pass

    @classmethod
    def bls12_381_serialize(cls, g: Any) -> bytes:
        """
        Serialize a bls12381 point.

        :param g: The point to be serialized.
        :type g: bytes
        :return: the serialized point
        :rtype: bytes
        """
        pass



from typing import Any

from boa3.internal.model.type.collection.sequence.uint256type import UInt256Type
from boa3.internal.model.type.itype import IType


class BlockHashType(UInt256Type):
    """
    A class used to indicate that a parameter or return on the manifest is a BlockHash. It's a subclass of UInt256Type.
    """

    def __init__(self):
        super().__init__()
        self._identifier = 'BlockHash'

    @classmethod
    def build(cls, value: Any = None) -> IType:
        return _BlockHash


_BlockHash = BlockHashType()



from typing import Any, Set

from boa3.internal.model.type.collection.mapping.mutable.mutablemappingtype import MutableMappingType
from boa3.internal.model.type.itype import IType


class DictType(MutableMappingType):
    """
    A class used to represent Python dict type
    """

    def __init__(self, keys_type: Set[IType] = None, values_type: Set[IType] = None):
        identifier = 'dict'
        keys_type = self.filter_types(keys_type)
        values_type = self.filter_types(values_type)
        super().__init__(identifier, keys_type, values_type)

    @property
    def default_value(self) -> Any:
        return dict()

    @classmethod
    def _is_type_of(cls, value: Any):
        return type(value) in [dict, DictType]

    def _init_class_symbols(self):
        super()._init_class_symbols()

        from boa3.internal.model.builtin.builtin import Builtin

        instance_methods = [Builtin.Copy,
                            Builtin.DictKeys,
                            Builtin.DictValues,
                            Builtin.DictPop,
                            Builtin.DictPopDefault,
                            ]

        for instance_method in instance_methods:
            self._instance_methods[instance_method.raw_identifier] = instance_method.build(self)

    def __hash__(self):
        return hash(self.identifier)



__all__ = [
    'CallFlags',
    'Contract',
    'ContractManifest',
    'call_contract',
    'create_contract',
    'update_contract',
    'destroy_contract',
    'get_minimum_deployment_fee',
    'get_call_flags',
    'create_standard_account',
    'create_multisig_account',
    'NEO',
    'GAS',
]


from typing import Any, List, Sequence

from boa3.builtin.interop.contract.callflagstype import CallFlags
from boa3.builtin.interop.contract.contract import Contract
from boa3.builtin.interop.contract.contractmanifest import ContractManifest
from boa3.builtin.type import ECPoint, UInt160


def call_contract(script_hash: UInt160, method: str, args: Sequence = (), call_flags: CallFlags = CallFlags.ALL) -> Any:
    """
    Calls a smart contract given the method and the arguments. Since the return is type Any, you'll probably need to
    type cast the return.

    >>> call_contract(NEO, 'balanceOf', [UInt160(b'\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2')])
    100

    :param script_hash: the target smart contract's script hash
    :type script_hash: UInt160
    :param method: the name of the method to be executed
    :type method: str
    :param args: the specified method's arguments
    :type args: Sequence[Any]
    :param call_flags: the CallFlags to be used to call the contract
    :type call_flags: CallFlags

    :return: the result of the specified method
    :rtype: Any

    :raise Exception: raised if there isn't a valid CallFlags, the script hash is not a valid smart contract or the
        method was not found or the arguments aren't valid to the specified method.
    """
    pass


def create_contract(nef_file: bytes, manifest: bytes, data: Any = None) -> Contract:
    """
    Creates a smart contract given the script and the manifest.

    >>> nef_file_ = get_script(); manifest_ = get_manifest()    # get the script and manifest somehow
    ... create_contract(nef_file_, manifest_, None)             # smart contract will be deployed
    {
        'id': 2,
        'update_counter': 0,
        'hash': b'\\x92\\x8f+1q\\x86z_@\\x94\\xf5pE\\xcb\\xb8 \\x0f\\\\`Z',
        'nef': b'NEF3neo3-boa by COZ-1.0.0\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x00\\x07W\\x00\\x02xy\\x9e@\\xf9\\7b\\xbb\\xcc',
        'manifest': {
            'name': 'TestContract',
            'group': [],
            'supported_standards': [],
            'abi': [[['test', [['a', 17], ['b', 17]], 17, 0, False]], []],
            'permissions': [],
            'trusts': [],
            'extras': 'null'
        },
    }

    :param nef_file: the target smart contract's compiled nef
    :type nef_file: bytes
    :param manifest: the manifest.json that describes how the script should behave
    :type manifest: bytes
    :param data: the parameters for the _deploy function
    :type data: Any

    :return: the contract that was created
    :rtype: Contract

    :raise Exception: raised if the nef or the manifest are not a valid smart contract.
    """
    pass


def update_contract(nef_file: bytes, manifest: bytes, data: Any = None):
    """
    Updates the executing smart contract given the script and the manifest.

    >>> nef_file_ = get_script(); manifest_ = get_manifest()    # get the script and manifest somehow
    ... update_contract(nef_file_, manifest_, None)             # smart contract will be updated
    None

    :param nef_file: the new smart contract's compiled nef
    :type nef_file: bytes
    :param manifest: the new smart contract's manifest
    :type manifest: bytes
    :param data: the parameters for the _deploy function
    :type data: Any

    :raise Exception: raised if the nef and the manifest are not a valid smart contract or the new contract is the
        same as the old one.
    """
    pass


def destroy_contract():
    """
    Destroy the executing smart contract.

    >>> destroy_contract()
    None

    """
    pass


def get_minimum_deployment_fee() -> int:
    """
    Gets the minimum fee of contract deployment.

    >>> get_minimum_deployment_fee()
    1000000000

    :return: the minimum fee of contract deployment
    """
    pass


def get_call_flags() -> CallFlags:
    """
    Gets the CallFlags in the current context.

    >>> get_call_flags()
    CallFlags.READ_ONLY
    """
    pass


def create_standard_account(pub_key: ECPoint) -> UInt160:
    """
    Calculates the script hash from a public key.

    >>> create_standard_account(ECPoint(b'\\x03\\x5a\\x92\\x8f\\x20\\x16\\x39\\x20\\x4e\\x06\\xb4\\x36\\x8b\\x1a\\x93\\x36\\x54\\x62\\xa8\\xeb\\xbf\\xf0\\xb8\\x81\\x81\\x51\\xb7\\x4f\\xaa\\xb3\\xa2\\xb6\\x1a'))
    b'\\r\\xa9g\\xa4\\x00C+\\xf2\\x7f\\x8e\\x8e\\xb4o\\xe8\\xace\\x9e\\xcc\\xde\\x04'

    :param pub_key: the given public key
    :type pub_key: ECPoint

    :return: the corresponding script hash of the public key
    :rtype: UInt160
    """
    pass


def create_multisig_account(m: int, pub_keys: List[ECPoint]) -> UInt160:
    """
    Calculates corresponding multisig account script hash for the given public keys.

    >>> create_multisig_account(1, [ECPoint(b'\\x03\\x5a\\x92\\x8f\\x20\\x16\\x39\\x20\\x4e\\x06\\xb4\\x36\\x8b\\x1a\\x93\\x36\\x54\\x62\\xa8\\xeb\\xbf\\xf0\\xb8\\x81\\x81\\x51\\xb7\\x4f\\xaa\\xb3\\xa2\\xb6\\x1a')])
    b'"5,\\xd2\\x9e\\xe7\\xb4\\x02\\x08b\\xdbd\\x1e\\xedx\\x82\\x8fU(m'

    :param m: the minimum number of correct signatures need to be provided in order for the verification to pass.
    :type m: int
    :param pub_keys: the public keys of the account
    :type pub_keys: List[ECPoint]

    :return: the hash of the corresponding account
    :rtype: UInt160
    """
    pass


NEO: UInt160 = UInt160()
"""
NEO's token script hash.

>>> NEO
b'\\xf5c\\xea@\\xbc(=M\\x0e\\x05\\xc4\\x8e\\xa3\\x05\\xb3\\xf2\\xa0s@\\xef'

:meta hide-value:
"""

GAS: UInt160 = UInt160()
"""
GAS' token script hash.

>>> GAS
b'\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2'

:meta hide-value:
"""



from __future__ import annotations

import ast
from abc import ABC
from typing import Dict, List, Optional, Tuple

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable


class ContractGetHashMethod(IBuiltinMethod, ABC):
    def __init__(self, script_hash: bytes, identifier: str, args: Dict[str, Variable] = None,
                 defaults: List[ast.AST] = None, return_type: IType = None,
                 vararg: Optional[Tuple[str, Variable]] = None,
                 kwargs: Optional[Dict[str, Variable]] = None):
        self._script_hash = script_hash
        super().__init__(identifier, args, defaults, return_type, vararg, kwargs)

    @property
    def script_hash(self) -> bytes:
        return self._script_hash

    def generate_internal_opcodes(self, code_generator):
        code_generator.convert_literal(self.script_hash)



import base58


def address_to_script_hash(address: str, account_version: int = None) -> bytes:
    data = base58.b58decode_check(address)
    if isinstance(account_version, int):
        assert data[0] == account_version, "Address version doesn't match"

    return data[1:]  # first position is address version


def address_from_script_hash(script_hash: bytes, account_version: int) -> str:
    data = bytearray(script_hash)
    data.insert(0, account_version)

    return base58.b58encode_check(data).decode(encoding='utf-8')



from typing import List, Union

from boa3_test.test_drive.model.smart_contract.testcontract import TestContract
from boa3_test.test_drive.testrunner.blockchain.storage import TestRunnerStorage, StorageKey


class StorageCollection:
    def __init__(self):
        self._internal_list: List[TestRunnerStorage] = []
        self._storage_contracts: List[TestContract] = []

    def append(self, new_storage: TestRunnerStorage):
        self._storage_contracts.append(new_storage.contract)
        return self._internal_list.append(new_storage)

    def remove(self, storage: TestRunnerStorage):
        try:
            index = self._internal_list.index(storage)
            self._storage_contracts.pop(index)
            self._internal_list.pop(index)
        except BaseException as e:
            raise e  # try except just to ensure that the value will be removed from both lists

    def get(self, storage_contract: TestContract, key: Union[bytes, str]):
        if isinstance(key, str):
            try:
                key = bytes(key)
            except TypeError:
                from boa3.internal import constants
                key = key.encode(encoding=constants.ENCODING)

        index = self._storage_contracts.index(storage_contract)
        try:
            storage = self._internal_list[index].values
            storage_key = StorageKey(key)

            return storage[storage_key]
        except:
            return None

    def clear(self):
        self._storage_contracts.clear()
        return self._internal_list.clear()

    def __len__(self):
        return len(self._internal_list)

    def __contains__(self, item) -> bool:
        return any(storage.is_valid_identifier(item) for storage in self._internal_list)

    def __getitem__(self, item) -> TestRunnerStorage:
        return next((storage for storage in self._internal_list if storage.is_valid_identifier(item)), None)

    def __str__(self):
        dict_repr = {}
        for storage in self._internal_list:
            contract_id = storage.contract
            if contract_id is not None:
                contract_id = contract_id.name if hasattr(contract_id, 'name') else contract_id.script_hash

            dict_repr[contract_id] = storage.values
        return str(dict_repr)

    def __repr__(self):
        return str(self)



from typing import Dict, Optional

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class AbsMethod(IBuiltinMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        identifier = 'abs'
        args: Dict[str, Variable] = {'val': Variable(Type.int)}
        super().__init__(identifier, args, return_type=Type.int)

    def generate_internal_opcodes(self, code_generator):
        code_generator.insert_opcode(Opcode.ABS)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None



import ast
from typing import Dict, List, Optional

from boa3.internal.compiler.compiledmetadata import CompiledMetadata
from boa3.internal.model.builtin.interop.contractgethashmethod import ContractGetHashMethod
from boa3.internal.model.builtin.interop.interopmethod import InteropMethod
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable


class NativeContractMethod(InteropMethod):

    def __init__(self, native_contract_script_hash_method: ContractGetHashMethod, identifier: str, syscall: str,
                 args: Dict[str, Variable] = None, defaults: List[ast.AST] = None, return_type: IType = None,
                 internal_call_args: int = None):
        super().__init__(identifier, syscall, args, defaults, return_type)
        self.script_hash_method = native_contract_script_hash_method
        min_no_args = len(self.args_without_default)
        if not isinstance(internal_call_args, int):
            internal_call_args = len(self.args)
        elif len(self.args_without_default) > internal_call_args:
            internal_call_args = min_no_args

        self.internal_call_args = internal_call_args

        from boa3.internal.neo3.contracts.contracttypes import CallFlags
        self._call_flags_default = CallFlags.ALL
        self._added_to_permissions = False
        self._pack_arguments = None   # defined during compilation
        self._method_token_id = None  # defined during compilation
        self.external_name = self._sys_call

    @property
    def contract_script_hash(self) -> bytes:
        return self.script_hash_method.script_hash

    def reset(self):
        # reset the object state to ensure the correct output when calling consecutive compilations
        super().reset()
        self._added_to_permissions = False
        self._pack_arguments = None
        self._method_token_id = None

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.interop.interop import Interop
        from boa3.internal.model.type.type import Type

        self._add_to_contract_permissions()

        call_flag = self._call_flags_default
        self_method_token_id = self._get_method_token_id(call_flag)

        if isinstance(self_method_token_id, int) and self_method_token_id >= 0:
            code_generator.convert_method_token_call(self_method_token_id)
        else:
            if self._pack_arguments is None:
                self._pack_arguments = False

            code_generator.convert_new_array(len(self.args))
            code_generator.convert_literal(call_flag)
            code_generator.convert_literal(self.method_name)
            code_generator.convert_builtin_method_call(self.script_hash_method, is_internal=True)
            code_generator.convert_builtin_method_call(Interop.CallContract, is_internal=True)

            if self.return_type is Type.none:
                code_generator.remove_stack_top_item()

    def _get_method_token_id(self, call_flag=None) -> Optional[int]:
        if self._method_token_id is None:
            from boa3.internal.compiler.codegenerator.vmcodemapping import VMCodeMapping

            if call_flag is None:
                call_flag = self._call_flags_default
            self._method_token_id = VMCodeMapping.instance().add_method_token(self, call_flag)
        return self._method_token_id

    def _add_to_contract_permissions(self):
        if not self._added_to_permissions:
            CompiledMetadata.instance().add_contract_permission(self.contract_script_hash, self._sys_call)
            self._added_to_permissions = True

    @property
    def pack_arguments(self) -> bool:
        if self._pack_arguments is None:
            from boa3.internal.compiler.codegenerator.vmcodemapping import VMCodeMapping
            self_method_token_id = self._get_method_token_id(self._call_flags_default)

            self._pack_arguments = not isinstance(self_method_token_id, int) or self_method_token_id < 0
        return self._pack_arguments

    @property
    def method_name(self) -> str:
        return self._sys_call



from typing import Any, Union

from boa3_test.test_drive.model.wallet.account import Account
from boa3_test.test_drive.neoxp.command import neoexpresscommand as neoxp
from boa3_test.test_drive.neoxp.command import neoxp_contract as contract
from boa3_test.test_drive.neoxp.command import neoxp_oracle as oracle
from boa3_test.test_drive.neoxp.command.neoexpresscommand import NeoExpressCommand


def create_checkpoint(checkpoint_path: str, force: bool = False) -> NeoExpressCommand:
    return neoxp.checkpoint.CheckpointCreateCommand(checkpoint_path, force=force)


def reset() -> NeoExpressCommand:
    return neoxp.reset.ResetCommand(force=True)


def fastfwd(block_count: int, time_interval_in_secs: int = 0) -> NeoExpressCommand:
    if not isinstance(block_count, int) or block_count < 1:
        block_count = 1

    return neoxp.fastforward.FastForwardCommand(block_count, time_interval_in_secs)


def transfer(sender: Account, receiver: Account, asset: str,
             quantity: Union[int, float], decimals: int = 0,
             data: Any = None) -> NeoExpressCommand:

    return neoxp.transfer.TransferAssetCommand(asset, sender, receiver, quantity,
                                               decimals=decimals,
                                               data=data)



import ast
from typing import Dict, List

from boa3.internal.model.builtin.interop.nativecontract import StdLibMethod
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class StrSplitMethod(StdLibMethod):
    def __init__(self):
        from boa3.internal.model.type.type import Type
        identifier = 'split'
        syscall = 'stringSplit'
        args: Dict[str, Variable] = {
            'self': Variable(Type.str),
            'sep': Variable(Type.str),
            'maxsplit': Variable(Type.int)
        }
        # whitespace is the default separator
        separator_default = ast.parse("' '").body[0].value
        # maxsplit the default value is -1
        maxsplit_default = ast.parse("-1").body[0].value.operand
        maxsplit_default.n = -1

        neo_internal_args = {
            'str': Variable(Type.str),
            'separator': Variable(Type.str),
            # TODO: include removeEmptyEntries
        }

        super().__init__(identifier, syscall, args, defaults=[separator_default, maxsplit_default],
                         return_type=Type.list.build_collection(Type.str),
                         internal_call_args=len(neo_internal_args))

    @property
    def generation_order(self) -> List[int]:
        # the original string must be the top value in the stack
        indexes = list(range(len(self.args)))
        str_index = list(self.args).index('self')

        if indexes[-1] != str_index:
            # context must be the last generated argument
            indexes.remove(str_index)
            indexes.append(str_index)
        return indexes

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp

        code_generator.duplicate_stack_item(3)
        code_generator.swap_reverse_stack_items(2)
        super().generate_internal_opcodes(code_generator)

        # if maxsplit > 0
        code_generator.duplicate_stack_item(2)
        code_generator.convert_literal(0)
        is_valid_split_count = code_generator.convert_begin_if()
        code_generator.change_jump(is_valid_split_count, Opcode.JMPLT)

        #   while len(array) < maxsplit + 1:
        while_start = code_generator.convert_begin_while()
        code_generator.duplicate_stack_top_item()

        #       concat values
        code_generator.duplicate_stack_item(4)
        code_generator.duplicate_stack_item(2)
        code_generator.insert_opcode(Opcode.POPITEM, pop_from_stack=True)
        code_generator.convert_operation(BinaryOp.Concat, is_internal=True)
        code_generator.duplicate_stack_item(2)
        code_generator.insert_opcode(Opcode.POPITEM, pop_from_stack=True)
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_operation(BinaryOp.Concat, is_internal=True)
        code_generator.convert_builtin_method_call(Builtin.SequenceAppend, is_internal=True)

        while_condition = code_generator.bytecode_size
        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.duplicate_stack_item(3)
        code_generator.insert_opcode(Opcode.INC)
        code_generator.convert_operation(BinaryOp.Gt, is_internal=True)

        code_generator.convert_end_while(while_start, while_condition, is_internal=True)

        code_generator.convert_end_if(is_valid_split_count, is_internal=True)
        # clean stack

        code_generator.swap_reverse_stack_items(3)
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()



from typing import Dict, Optional, Any, Sequence

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class AbortMethod(IBuiltinMethod):

    def __init__(self, message_type=None):
        import ast
        from boa3.internal.model.type.itype import IType
        from boa3.internal.model.type.type import Type

        identifier = 'abort'

        if isinstance(message_type, IType):
            args: Dict[str, Variable] = {
                'msg': Variable(Type.optional.build(message_type))
            }
            defaults = [
                ast.parse(f'{Type.none.default_value}').body[0].value
            ]

        else:
            args: Dict[str, Variable] = {}
            defaults = []

        super().__init__(identifier, args, defaults, return_type=Type.none)

    @property
    def identifier(self) -> str:
        if 'msg' not in self.args:
            return super().identifier

        return '-{0}_{1}'.format(self._identifier, self.args['msg'].type.identifier)

    def generate_internal_opcodes(self, code_generator):
        code_generator.insert_opcode(Opcode.ABORT)

    @property
    def _args_on_stack(self) -> int:
        return 0

    def build(self, value: Any):
        if not isinstance(value, Sequence):
            value = [value]

        if len(value) == len(self.args):
            return self

        if len(value) > 0:
            obj = AbortStrMethod()
        else:
            from boa3.internal.model.builtin.builtin import Builtin
            obj = Builtin.Abort
        return obj

    @property
    def _body(self) -> Optional[str]:
        return None


class AbortStrMethod(AbortMethod):
    def __init__(self):
        from boa3.internal.model.type.type import Type
        message_type = Type.str
        super().__init__(message_type)

    def msg_arg_type(self) -> IType:
        from boa3.internal.model.type.type import Type
        return self.args['msg'].type if 'msg' in self.args else Type.optional.build(Type.str)

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.type.type import Type

        msg = code_generator._stack[-1] if len(code_generator._stack) else self.msg_arg_type()
        # if msg is optional[str] both variables are going to be set as True
        is_not_str = not Type.str.is_type_of(msg)
        is_not_none = not Type.none.is_type_of(msg)

        if is_not_str and is_not_none:
            # need to check the type at runtime
            code_generator.duplicate_stack_top_item()
            code_generator.insert_type_check(Type.str.stack_item)

            check_valid = code_generator.convert_begin_if()
            code_generator.change_jump(check_valid, Opcode.JMPIF)
            super().generate_internal_opcodes(code_generator)
            code_generator.convert_end_if(check_valid, is_internal=True)
            type_check_done = True
        else:
            type_check_done = False

        code_generator.convert_abort(has_message=True, is_internal=type_check_done)



__all__ = [
    'NeoExpressWrapper',
    '_NEOXP_CONFIG_LOCK'
]

import os.path
import re
import subprocess
import threading
from typing import Tuple, List, Union

from boa3.internal.neo3.core.types import UInt256
from boa3_test.test_drive.neoxp.command import neoexpresscommand as neoxp
from boa3_test.test_drive.neoxp.model.neoxpconfig import NeoExpressConfig
from boa3_test.test_drive.testrunner.blockchain.block import TestRunnerBlock as Block
from boa3_test.test_drive.testrunner.blockchain.contract import TestRunnerContract as Contract
from boa3_test.test_drive.testrunner.blockchain.transaction import TestRunnerTransaction as Transaction
from boa3_test.test_drive.testrunner.blockchain.transactionlog import TestRunnerTransactionLog as TransactionLog

_NEOXP_CONFIG_LOCK = threading.Lock()


class NeoExpressWrapper:
    @classmethod
    def run_neo_express_cli(cls, command: neoxp.NeoExpressCommand) -> Tuple[str, str]:
        neoxp_args = ['neoxp']
        neoxp_args.extend(command.cli_command().split())

        cli_result = cls._run_cli(*neoxp_args)
        return cli_result

    @classmethod
    def _run_cli(cls, *neoxp_args: str):
        process = subprocess.Popen(neoxp_args,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.STDOUT,
                                   text=True)
        return process.communicate()

    @staticmethod
    def get_config_data(neoxp_path: str) -> NeoExpressConfig:
        return NeoExpressConfig(neoxp_path)

    @classmethod
    def create_neo_express_instance(cls, neoxp_path: str) -> str:
        command = neoxp.create.CreateCommand(config_output=neoxp_path,
                                             node_count=1,
                                             force=True)
        with _NEOXP_CONFIG_LOCK:
            stdout, stderr = cls.run_neo_express_cli(command)

        return stdout

    @classmethod
    def reset_neo_express_instance(cls, neoxp_path: str, check_point_file: str = None) -> str:
        if isinstance(check_point_file, str) and os.path.isfile(check_point_file):
            command = neoxp.checkpoint.CheckpointRestoreCommand(neo_express_data_file=neoxp_path,
                                                                file_name=check_point_file,
                                                                force=True)
        else:
            command = neoxp.reset.ResetCommand(neo_express_data_file=neoxp_path,
                                               force=True)
        stdout, stderr = cls.run_neo_express_cli(command)
        return stdout

    @classmethod
    def get_deployed_contracts(cls, neoxp_path: str, check_point_file: str = None) -> List[Contract]:
        command = neoxp.contract.ContractListCommand(neo_express_data_file=neoxp_path)
        with _NEOXP_CONFIG_LOCK:
            cls.reset_neo_express_instance(neoxp_path, check_point_file)
            stdout, stderr = cls.run_neo_express_cli(command)

        contracts = []
        for line in stdout.splitlines():
            try:
                groups = re.match(r'(?P<name>.*?) \((?P<scripthash>0x\w+)\)', line).groupdict()
                contract_name = groups['name']
                contract_hash = groups['scripthash']
                found_contract = Contract(contract_name, contract_hash)
                contracts.append(found_contract)
            except:
                # don't break if some line doesn't match the regex
                continue
        return contracts

    @classmethod
    def _get_transaction_raw(cls, neoxp_path: str, tx_hash: UInt256, check_point_file: str = None) -> str:
        command = neoxp.show.ShowTransactionCommand(tx_hash.to_array(), neo_express_data_file=neoxp_path)
        with _NEOXP_CONFIG_LOCK:
            if isinstance(check_point_file, str):
                cls.reset_neo_express_instance(neoxp_path, check_point_file)
            stdout, stderr = cls.run_neo_express_cli(command)

        return stdout

    @classmethod
    def get_transaction(cls, neoxp_path: str, tx_hash: UInt256, check_point_file: str = None) -> Transaction:
        raw_result = cls._get_transaction_raw(neoxp_path, tx_hash, check_point_file)

        tx: Transaction
        try:
            import json
            result_json = json.loads(raw_result)
            tx = Transaction.from_json(result_json['transaction'], neoxp_config=cls.get_config_data(neoxp_path))
        except:
            tx = None

        return tx

    @classmethod
    def get_transaction_log(cls, neoxp_path: str, tx_hash: UInt256, check_point_file: str = None, contract_collection=None) -> TransactionLog:
        raw_result = cls._get_transaction_raw(neoxp_path, tx_hash, check_point_file)

        tx_log: TransactionLog
        try:
            import json
            result_json = json.loads(raw_result)
            tx_log = TransactionLog.from_json(result_json['application-log'], contract_collection)
            tx_log._tx_id = tx_hash
        except:
            tx_log = None

        return tx_log

    @classmethod
    def get_latest_block(cls, neoxp_path: str, check_point_file: str = None) -> Block:
        return cls.get_block(neoxp_path, check_point_file=check_point_file)

    @classmethod
    def get_block(cls, neoxp_path: str, block_hash_or_index: Union[UInt256, int] = None,
                  check_point_file: str = None) -> Block:
        if isinstance(block_hash_or_index, (UInt256, int)):
            if isinstance(block_hash_or_index, int):
                if block_hash_or_index < 0:
                    block_hash_or_index = 0
            else:
                block_hash_or_index = block_hash_or_index.to_array()

            command = neoxp.show.ShowBlockCommand(block_hash_or_index,
                                                  neo_express_data_file=neoxp_path)
        else:
            command = neoxp.show.ShowBlockCommand(neo_express_data_file=neoxp_path)

        with _NEOXP_CONFIG_LOCK:
            if isinstance(check_point_file, str):
                cls.reset_neo_express_instance(neoxp_path, check_point_file)
            stdout, stderr = cls.run_neo_express_cli(command)

        block: Block
        try:
            import json
            result_json = json.loads(stdout)
            block = Block.from_json(result_json, neoxp_config=cls.get_config_data(neoxp_path))
        except:
            block = None

        return block

    @classmethod
    def run_batch(cls, neoxp_path: str, batch_path: str, reset: bool = False, check_point_file: str = None) -> str:
        if isinstance(check_point_file, str) and not os.path.isfile(check_point_file):
            check_point_file = None
        command = neoxp.batch.BatchCommand(batch_path,
                                           neo_express_data_file=neoxp_path,
                                           check_point_file=check_point_file,
                                           reset=reset)
        with _NEOXP_CONFIG_LOCK:
            stdout, stderr = cls.run_neo_express_cli(command)

        return stdout

    @classmethod
    def oracle_response(cls, neoxp_path: str, url: str, response_path: str, request_id: int = None,
                        check_point_file: str = None) -> List[UInt256]:
        command = neoxp.oracle.OracleResponseCommand(url, response_path, request_id, neo_express_data_file=neoxp_path)

        with _NEOXP_CONFIG_LOCK:
            if isinstance(check_point_file, str):
                cls.reset_neo_express_instance(neoxp_path, check_point_file)
            stdout, stderr = cls.run_neo_express_cli(command)

        import re
        from boa3.internal.neo import from_hex_str

        oracle_response_submitted = re.findall(r"0x\w{64}", stdout)

        return [UInt256(from_hex_str(tx)) for tx in oracle_response_submitted]



def create_custom_id(value: str, use_time: bool = True) -> str:
    import platform

    suffix = platform.python_version()
    if use_time:
        import time
        suffix = f'{time.time_ns()}_{suffix}'

    if not isinstance(value, str) or len(value) == 0 or value.isspace():
        return suffix

    return f'{value}_{suffix}'



from __future__ import annotations

__all__ = [
    'Event',
    'UInt160',
    'UInt256',
    'ECPoint',
    'Address',
    'BlockHash',
    'PublicKey',
    'ScriptHash',
    'ScriptHashLittleEndian',
    'TransactionId',
]

from typing import Union


class Event:
    """
    Describes an action that happened in the blockchain.
    Neo3-Boa compiler won't recognize the `__init__` of this class. To create a new Event, use the method `CreateNewEvent`:

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/develop/write/basics#events>`__ to learn more
    about Events.

    >>> from boa3.builtin.compile_time import CreateNewEvent
    ... new_event: Event = CreateNewEvent(  # create a new Event with the CreateNewEvent method
    ...     [
    ...        ('name', str),
    ...        ('amount', int)
    ...     ],
    ...     'New Event'
    ... )
    """

    def __call__(self, *args, **kwargs):
        pass


class UInt160(bytes):
    """
    Represents a 160-bit unsigned integer.
    """

    def __init__(self, arg: Union[bytes, int] = 0):
        super().__init__()
        pass


class UInt256(bytes):
    """
    Represents a 256-bit unsigned integer.
    """

    def __init__(self, arg: Union[bytes, int] = 0):
        super().__init__()
        pass


class ECPoint(bytes):
    """
    Represents a coordinate pair for elliptic curve cryptography (ECC) structures.
    """

    def __init__(self, arg: bytes):
        super().__init__()
        pass

    def to_script_hash(self) -> bytes:
        """
        Converts a data to a script hash.

        :return: the script hash of the data
        :rtype: bytes
        """
        pass


Address = str
"""
A type used only to indicate that a parameter or return on the manifest should be treated as an Address.
Same as str.

:meta hide-value:
"""


BlockHash = UInt256
"""
A type used only to indicate that a parameter or return on the manifest should be treated as a BlockHash.
Same as UInt256.

:meta hide-value:
"""


PublicKey = ECPoint
"""
A type used only to indicate that a parameter or return on the manifest should be treated as a PublicKey.
Same as ECPoint.

:meta hide-value:
"""


ScriptHash = UInt160
"""
A type used only to indicate that a parameter or return on the manifest should be treated as a ScriptHash.
Same as UInt160.

:meta hide-value:
"""


ScriptHashLittleEndian = UInt160
"""
A type used only to indicate that a parameter or return on the manifest should be treated as a
ScriptHashLittleEndian.
Same as UInt160.

:meta hide-value:
"""


TransactionId = UInt256
"""
A type used only to indicate that a parameter or return on the manifest should be treated as a TransactionId.
Same as UInt256.

:meta hide-value:
"""



from typing import Any, Dict, List, Optional, Union

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class IsInstanceMethod(IBuiltinMethod):

    def __init__(self, target_type: IType = None):
        from boa3.internal.model.type.type import Type
        identifier = 'isinstance'

        args: Dict[str, Variable] = {
            'x': Variable(Type.any),
            'A_tuple': None
        }

        super().__init__(identifier, args, return_type=Type.bool)

        from boa3.internal.model.type.collection.sequence.tupletype import TupleType
        from boa3.internal.model.type.annotation.metatype import MetaType
        from boa3.internal.model.type.annotation.uniontype import UnionType

        if not isinstance(target_type, IType):
            instances = [Type.none]
        elif isinstance(target_type, TupleType) and isinstance(target_type.item_type, UnionType):
            instances = [typ.meta_type if isinstance(typ, MetaType) else typ
                         for typ in target_type.item_type.union_types]
        else:
            instances = [target_type.meta_type if isinstance(target_type, MetaType) else target_type]

        self._instances_type: List[IType] = instances

    def set_instance_type(self, value: List[IType]):
        new_list = []
        for tpe in value:
            if isinstance(tpe, IType):
                if not any(tpe.raw_identifier == other.raw_identifier for other in new_list):
                    new_list.append(tpe)

        self._instances_type = new_list

    @property
    def identifier(self) -> str:
        from boa3.internal.model.type.type import Type
        if (len(self._instances_type) == 0
                or (len(self._instances_type) == 1 and self._instances_type[0] in (None, Type.none))
            ):
            return self._identifier

        types = list({tpe.raw_identifier for tpe in self._instances_type})
        types.sort()
        return '-{0}_of_{1}'.format(self._identifier, '_or_'.join(types))

    def args_to_be_generated(self) -> List[int]:
        args = [name for name, symbol in self.args.items() if isinstance(symbol, Variable)]
        return [list(self.args).index(key) for key in args]

    @property
    def is_supported(self) -> bool:
        from boa3.internal.model.type.classes.classtype import ClassType
        return not any(isinstance(param, ClassType) and len(param.is_instance_opcodes()) == 0
                       for param in self._instances_type)

    def not_supported_str(self, callable_id: str) -> str:
        types = (self._instances_type[0].identifier if len(self._instances_type) == 1
                 else '({0})'.format(', '.join([arg.identifier for arg in self._instances_type])))

        return '{0}({1}, {2})'.format(callable_id, self.arg_x, types)

    @property
    def arg_x(self) -> Variable:
        return self.args['x']

    def validate_parameters(self, *params: Union[IExpression, IType]) -> bool:
        if len(params) != 2:
            return False

        return not any(not isinstance(param, (IExpression, IType)) for param in params)

    def generate_internal_opcodes(self, code_generator):
        if len(self._instances_type) == 0:
            code_generator.insert_type_check(None)
            return

        ifs = []
        types = self._instances_type.copy()
        for type_to_check in types[:-1]:
            code_generator.duplicate_stack_top_item()
            type_to_check.generate_is_instance_type_check(code_generator)

            if_is_instance = code_generator.convert_begin_if()
            code_generator.change_jump(if_is_instance, Opcode.JMPIF)

            ifs.append(if_is_instance)

        types[-1].generate_is_instance_type_check(code_generator)
        if len(types) > 1:
            last_if = code_generator.convert_begin_if()
            code_generator.change_jump(last_if, Opcode.JMP)

        for begin_if in ifs:
            code_generator.convert_end_if(begin_if, is_internal=True)

        if len(types) > 1:
            code_generator.remove_stack_top_item()
            code_generator.convert_literal(True)
            code_generator.convert_end_if(last_if, is_internal=True)

    @property
    def _args_on_stack(self) -> int:
        return 1

    @property
    def generation_order(self) -> List[int]:
        # type should not be converted
        indexes = super().generation_order
        typ_index = list(self.args).index('A_tuple')

        if typ_index in indexes:
            indexes.remove(typ_index)

        return indexes

    @property
    def _body(self) -> Optional[str]:
        return

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, list) and self.validate_parameters(*value):
            return IsInstanceMethod(value[-1])
        return super().build(value)



__all__ = [
    'BoaNeoExpressBatch'
]


from boa3_test.test_drive.neoxp.batch import NeoExpressBatch
from boa3_test.tests.test_drive.neoxp import utils as boa_neoxp_utils


class BoaNeoExpressBatch(NeoExpressBatch):
    def _run_batch(self, neoxp_path: str, batch_file_path: str, reset: bool = False, check_point_file: str = None):
        return boa_neoxp_utils.run_batch(neoxp_path, batch_file_path, reset=reset, check_point_file=check_point_file)
